import torch
import numpy as np
import seaborn as sb
import networkx as nx
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches

def plot_series(x : np.array, columns : list):
    """
    Plots time series data from an array using Seaborn's line plot functionality. Each series
    is represented as a separate line on the plot.

    Parameters
    ----------
    x : np.array
        A NumPy array containing the data points of the time series.
    columns : list
        List of strings representing the column names or labels for each time series in the array.

    Returns
    -------
    None
        Displays the plot directly and returns None.
    """
    data = dict.fromkeys(columns)
    for i ,s in enumerate(columns):
        data[s] = x[:, i]
    sb.relplot(data, kind = "line", legend = columns)

def plot_graph(states: np.array, graph : np.array, classes=None, layout=None):
    """
    Plots a graph using NetworkX where nodes are colored based on their state and annotated with class labels.
    The graph's layout can be specified; otherwise, a spring layout is used.

    Parameters
    ----------
    states : np.array
        Array containing the state of each node used to determine the color of nodes.
    graph : np.array
        A 2D array where each column represents an edge between two nodes (start and end).
    classes : list, optional
        List of class labels corresponding to each node. Default is None.
    layout : dict, optional
        A dictionary defining the position of each node for custom layouts. If None, a spring layout is used. Default is None.

    Returns
    -------
    dict
        The positions of the nodes in the plot, useful for customizing the layout or for further graphical analysis.
    """
    graph = [(graph[0, i], graph[1, i]) for i in range(len(graph[0]))]

    g = nx.Graph()
    g.add_nodes_from(range(len(states)))
    g.add_edges_from(graph)
    color_map = {
        0:'red',
        1:'blue',
        2:'green',
        3:'yellow',
        4:'purple'
    }
    colors = []
    labels = {}
    for node in states:
        colors.append(color_map[node])
        labels[node] = classes[node]

    plt.figure(figsize=(10,10))

    if layout is not None:
        nx.draw(g, pos=layout, node_color=colors, font_size=10, node_size=15)
    else:
        pos = nx.spring_layout(g, seed=42)
        nx.draw(g, pos = pos, node_color=colors, font_size=10, node_size=25)

    patches = []
    for i, item in enumerate(color_map.items()):
        if i>=len(classes):
            break
        patches.append(mpatches.Patch(color=item[1], label=classes[item[0]]))


    plt.legend(handles=patches)
    plt.show()

    return pos



# # this code is adopted from https://github.com/dolphin-zs/HierST/blob/main/src/hierst.py

# import numpy as np
# import torch
# import torch.nn as nn
# import torch.nn.functional as F

# import torch_geometric.nn as PyG

# from .base import BaseModel


# #--------------------------------For Temporal Module: NBeats------------------------------

# class Block(nn.Module):

#     def __init__(self, hidden_channels, thetas_dim, backcast_length=10, forecast_length=5,share_thetas=False,
#                  nb_harmonics=None):
#         super(Block, self).__init__()
#         self.hidden_channels = hidden_channels
#         self.thetas_dim = thetas_dim
#         self.backcast_length = backcast_length
#         self.forecast_length = forecast_length
#         self.share_thetas = share_thetas

#         # (batch_size, hidden_channels, backcast_length)
#         self.conv1d_1 = nn.Conv1d(hidden_channels, hidden_channels, kernel_size=3, stride=1, dilation=1, padding=1)   # zero_padding to align length
#         self.conv1d_2 = nn.Conv1d(hidden_channels, hidden_channels, kernel_size=3, stride=1, dilation=1, padding=1)
#         self.conv1d_3 = nn.Conv1d(hidden_channels, hidden_channels, kernel_size=3, stride=1, dilation=1, padding=1)
#         self.conv1d_4 = nn.Conv1d(hidden_channels, hidden_channels, kernel_size=3, stride=1, dilation=1, padding=1)

#         # (batch_size, hidden_channels, thetas_dim)
#         if share_thetas:
#             self.theta_f_fc = self.theta_b_fc = nn.Linear(backcast_length, thetas_dim)
#         else:
#             self.theta_b_fc = nn.Linear(backcast_length, thetas_dim)
#             self.theta_f_fc = nn.Linear(backcast_length, thetas_dim)


#         self.backcast_linspace, self.forecast_linspace = linspace(backcast_length, forecast_length)

#     def forward(self, x):
#         x = F.relu(self.conv1d_1(x))
#         x = F.relu(self.conv1d_2(x))
#         x = F.relu(self.conv1d_3(x))
#         x = F.relu(self.conv1d_4(x))
#         return x

# def seasonality_model(thetas, t):
#     p = thetas.size()[-1]
#     assert p <= thetas.shape[1], 'thetas_dim is too big.'
#     p1, p2 = (p // 2, p // 2) if p % 2 == 0 else (p // 2, p // 2 + 1)
#     s1 = torch.FloatTensor([np.cos(2 * np.pi * i * t) for i in range(p1)])  # H/2-1
#     s2 = torch.FloatTensor([np.sin(2 * np.pi * i * t) for i in range(p2)])
#     S = torch.cat([s1, s2]).to(thetas.get_device())
#     return thetas.matmul(S)


# def trend_model(thetas, t):
#     p = thetas.size()[-1]
#     assert p <= 4, 'thetas_dim is too big.'
#     T = torch.FloatTensor([t ** i for i in range(p)]).to(thetas.get_device())
#     return thetas.matmul(T)


# def linspace(backcast_length, forecast_length):
#     lin_space = np.linspace(-backcast_length, forecast_length, backcast_length + forecast_length, endpoint=False)
#     b_ls = lin_space[:backcast_length] / (backcast_length + forecast_length)
#     f_ls = lin_space[backcast_length:] / (backcast_length + forecast_length)
#     return b_ls, f_ls

# class SeasonalityBlock(Block):

#     def __init__(self, hidden_channels, thetas_dim, backcast_length=10, forecast_length=5):
#         super(SeasonalityBlock, self).__init__(hidden_channels, forecast_length, backcast_length,
#                                                    forecast_length, share_thetas=True)

#         self.forecast_pool = nn.AdaptiveMaxPool1d(1)
#         self.forecast_map = nn.Linear(hidden_channels, forecast_length)

#     def forward(self, x):
#         x = super(SeasonalityBlock, self).forward(x)
#         backcast = seasonality_model(self.theta_b_fc(x), self.backcast_linspace)   # (batch_size, hidden_channels, backcast_length)
#         forecast = seasonality_model(self.theta_f_fc(x), self.forecast_linspace)   # (batch_size, hidden_channels, forecast_length)
#         forecast = self.forecast_map(self.forecast_pool(forecast).squeeze())       # (batch_size, forecast_length)

#         return backcast, forecast


# class TrendBlock(Block):

#     def __init__(self, hidden_channels, thetas_dim,   backcast_length=10, forecast_length=5):
#         super(TrendBlock, self).__init__(hidden_channels, thetas_dim, backcast_length,
#                                          forecast_length, share_thetas=True)
#         self.forecast_pool = nn.AdaptiveMaxPool1d(1)
#         self.forecast_map = nn.Linear(hidden_channels, forecast_length)

#     def forward(self, x):
#         x = super(TrendBlock, self).forward(x)
#         backcast = trend_model(self.theta_b_fc(x), self.backcast_linspace)
#         forecast = trend_model(self.theta_f_fc(x), self.forecast_linspace)
#         forecast = self.forecast_map(self.forecast_pool(forecast).squeeze())       # (batch_size, forecast_length)

#         return backcast, forecast


# class GenericBlock(Block):

#     def __init__(self, hidden_channels, thetas_dim,  backcast_length=10, forecast_length=5):
#         super(GenericBlock, self).__init__(hidden_channels, thetas_dim, backcast_length, forecast_length)

#         self.backcast_fc = nn.Linear(thetas_dim, backcast_length)
#         self.forecast_fc = nn.Linear(thetas_dim, forecast_length)
#         self.forecast_pool = nn.AdaptiveMaxPool1d(1)
#         self.forecast_map = nn.Linear(hidden_channels, forecast_length)


#     def forward(self, x):
#         x = super(GenericBlock, self).forward(x)

#         theta_b = F.relu(self.theta_b_fc(x))
#         theta_f = F.relu(self.theta_f_fc(x))

#         backcast = self.backcast_fc(theta_b)
#         forecast = self.forecast_map(self.forecast_pool(F.relu(self.forecast_fc(theta_f))).squeeze())

#         return backcast, forecast


# class NBeatsNet(nn.Module):

#     SEASONALITY_BLOCK = 'seasonality'
#     TREND_BLOCK = 'trend'
#     GENERIC_BLOCK = 'generic'

#     def __init__(self,
#                  input_channels = 1,
#                  stack_types=(TREND_BLOCK, SEASONALITY_BLOCK),
#                  nb_blocks_per_stack=3,
#                  forecast_length=14,
#                  backcast_length=1,
#                  thetas_dims=(4, 8),
#                  share_weights_in_stack=False,
#                  hidden_channels=64):
#         super(NBeatsNet, self).__init__()
#         self.input_channels = input_channels
#         self.forecast_length = forecast_length
#         self.backcast_length = backcast_length
#         self.hidden_channels = hidden_channels
#         self.nb_blocks_per_stack = nb_blocks_per_stack
#         self.share_weights_in_stack = share_weights_in_stack
#         self.stack_types = stack_types
#         self.stacks = []
#         self.thetas_dim = thetas_dims
#         self.parameters = []
#         self.conv1d_flatten = nn.Conv1d(input_channels, hidden_channels, kernel_size=3, stride=1, dilation=1, padding=1)
#         self.parameters.extend(self.conv1d_flatten.parameters())
#         # print(f'| N-Beats')
#         # print(f'     | -- {self.conv1d_flatten}')
#         for stack_id in range(len(self.stack_types)):
#             self.stacks.append(self.create_stack(stack_id))
#         self.parameters = nn.ParameterList(self.parameters)

#     def create_stack(self, stack_id):
#         stack_type = self.stack_types[stack_id]
#         # print(f'| --  Stack {stack_type.title()} (#{stack_id}) (share_weights_in_stack={self.share_weights_in_stack})')
#         blocks = []
#         for block_id in range(self.nb_blocks_per_stack):
#             block_init = NBeatsNet.select_block(stack_type)
#             if self.share_weights_in_stack and block_id != 0:
#                 block = blocks[-1]  # pick up the last one when we share weights.
#             else:
#                 block = block_init(self.hidden_channels, self.thetas_dim[stack_id],
#                                    self.backcast_length, self.forecast_length)
#                 self.parameters.extend(block.parameters())
#             # print(f'     | -- {block}')
#             blocks.append(block)
#         return blocks

#     @staticmethod
#     def select_block(block_type):
#         if block_type == NBeatsNet.SEASONALITY_BLOCK:
#             return SeasonalityBlock
#         elif block_type == NBeatsNet.TREND_BLOCK:
#             return TrendBlock
#         else:
#             return GenericBlock

#     def forward(self, backcast):
#         forecast = torch.zeros(size=(backcast.size()[0], self.forecast_length)).to(backcast.get_device())  # maybe batch size here.
#         backcast = self.conv1d_flatten(backcast)     # (batch_size, input_channels, backcast_length) --> (batch_size, hidden_channels, backcast_length)
#         for stack_id in range(len(self.stacks)):
#             for block_id in range(len(self.stacks[stack_id])):
#                 b, f = self.stacks[stack_id][block_id](backcast)
#                 backcast = backcast - b
#                 forecast = forecast + f
#         return backcast, forecast

# class NBeatsEncoder(nn.Module):
#     def __init__(self, num_nodes, date_emb_dim, id_emb_dim, lookahead_days, lookback_days, hidden_dim, day_fea_dim, block_size):
#         super().__init__()
#         self.week_em = nn.Embedding(7, date_emb_dim)
#         self.id_em = nn.Embedding(num_nodes, id_emb_dim)
#         self.lookahead_days = lookahead_days
#         self.lookback_days = lookback_days
#         self.hidden_dim = hidden_dim

#         day_input_dim = day_fea_dim - 1 + self.week_em.embedding_dim + self.id_em.embedding_dim

#         self.day_n_beats = NBeatsNet(
#             input_channels=day_input_dim,
#             stack_types=('generic','generic'),
#             nb_blocks_per_stack=block_size,
#             forecast_length=lookahead_days,
#             backcast_length=lookback_days,
#             thetas_dims=(4,8),
#             share_weights_in_stack=True,
#             hidden_channels=hidden_dim)


#     def add_date_embed(self, input_day):
#         # last 1 dims correspond to weekday
#         x = input_day[:, :, :, :-1]
#         weekday = self.week_em(input_day[:, :, :, -1].long())
#         # return torch.cat([x, month, day, weekday], dim=-1)
#         return torch.cat([x, weekday], dim=-1)

#     def add_id_embed(self, input, g):
#         # add cent id index
#         sz = input.size()
#         cent_id = self.id_em(g['cent_n_id'].reshape(1,sz[1],1).expand(sz[0],sz[1],sz[2]).long())
#         return torch.cat([input, cent_id], dim=-1)

#     def permute_feature(self, x):
#         sz = x.size()
#         x = x.view(-1, sz[-2], sz[-1])
#         x = x.permute(0, 2, 1)
#         return x

#     def forward(self, input_day, g):
#         # [batch_size, node_num, seq_len, fea_dim]
#         sz = input_day.size()
#         # forecast_lr = self.linear_reg(input_day)
#         input_day = self.add_date_embed(input_day)
#         input_day = self.add_id_embed(input_day, g)
#         input_day = self.permute_feature(input_day)

#         backcast_day, forecast_day = self.day_n_beats(input_day)

#         backcast_day = backcast_day.view(sz[0], sz[1], self.hidden_dim, self.lookback_days)
#         forecast_day = forecast_day.squeeze().view(sz[0], sz[1], self.lookahead_days)

#         return backcast_day, forecast_day



# #------------------------------For Spatial Module----------------------


# class BaseGNNNet(nn.Module):
#     def __init__(self):
#         super().__init__()

#     def dataflow_forward(self, X, g):
#         raise NotImplementedError

#     def subgraph_forward(self, X, g):
#         raise NotImplementedError

#     def forward(self, X, g, **kwargs):
#         if g['type'] == 'dataflow':
#             return self.dataflow_forward(X, g, **kwargs)
#         elif g['type'] == 'subgraph':
#             return self.subgraph_forward(X, g, **kwargs)
#         else:
#             raise Exception('Unsupported graph type {}'.format(g['type']))
        


# class GCNConv(PyG.MessagePassing):
#     def __init__(self, gcn_in_dim, config, **kwargs):
#         super().__init__(aggr=config.gcn_aggr, node_dim=-2, **kwargs)

#         self.gcn_in_dim = gcn_in_dim
#         self.gcn_node_dim = config.gcn_node_dim
#         self.gcn_dim = config.gcn_dim

#         self.fea_map = nn.Linear(self.gcn_in_dim, self.gcn_dim)

#     def forward(self, x, edge_index, edge_weight):
#         batch_size, num_nodes, _ = x.shape
#         if edge_weight.shape[0] == x.shape[0]:
#             num_edges = edge_weight.shape[1]
#             edge_weight = edge_weight.reshape(batch_size, num_edges, 1)
#         else:
#             num_edges = edge_weight.shape[0]
#             edge_weight = edge_weight.reshape(1, num_edges, 1).expand(batch_size, -1, -1)

#         # Calculate type-aware node info
#         if isinstance(x, torch.Tensor):
#             x = self.fea_map(x)
#         else:
#             x = (self.fea_map(x[0]), self.fea_map(x[1]))

#         return self.propagate(edge_index, x=x, edge_weight=edge_weight)

#     def message(self, x_j, edge_weight):
#         return x_j * edge_weight

#     def update(self, aggr_out, x):
#         if not isinstance(x, torch.Tensor):
#             x = x[1]

#         return x + aggr_out


# class MyGATConv(PyG.MessagePassing):
#     def __init__(self, gcn_in_dim, config, **kwargs):
#         super().__init__(aggr=config.gcn_aggr, node_dim=-2, **kwargs)

#         self.gcn_in_dim = gcn_in_dim
#         self.gcn_node_dim = config.gcn_node_dim
#         self.gcn_dim = config.gcn_dim

#         self.fea_map = nn.Linear(self.gcn_in_dim, self.gcn_dim)
#         self.msg_map = nn.Linear(self.gcn_dim, self.gcn_dim)

#     def forward(self, x, edge_index, edge_weight):
#         batch_size, num_nodes, _ = x.shape
#         if edge_weight.shape[0] == x.shape[0]:
#             num_edges = edge_weight.shape[1]
#             edge_weight = edge_weight.reshape(batch_size, num_edges, 1)
#         else:
#             num_edges = edge_weight.shape[0]
#             edge_weight = edge_weight.reshape(1, num_edges, 1).expand(batch_size, -1, -1)

#         # Calculate type-aware node info
#         if isinstance(x, torch.Tensor):
#             x = self.fea_map(x)
#         else:
#             x = (self.fea_map(x[0]), self.fea_map(x[1]))

#         return self.propagate(edge_index, x=x, edge_weight=edge_weight)

#     def message(self, x_i, x_j, edge_weight):
#         x_i = self.msg_map(x_i)
#         x_j = self.msg_map(x_j)
#         gate = torch.sigmoid((x_i * x_j).sum(dim=-1, keepdim=True))
#         return x_j * edge_weight * gate

#     def update(self, aggr_out, x):
#         if not isinstance(x, torch.Tensor):
#             x = x[1]

#         return x + aggr_out

# class NodeGNN(BaseGNNNet):
#     def __init__(self, gcn_type, gcn_in_dim, gcn_layer_num):
#         super().__init__()

#         self.layer_num = gcn_layer_num
#         assert self.layer_num >= 1

#         if gcn_type == 'gcn':
#             GCNClass = GCNConv
#         elif gcn_type == 'gat':
#             GCNClass = MyGATConv
#         else:
#             raise Exception(f'Unsupported gcn_type {gcn_type}')

#         convs = [GCNClass(gcn_in_dim, )]
#         for _ in range(self.layer_num-1):
#             convs.append(GCNClass(.gcn_dim, ))
#         self.convs = nn.ModuleList(convs)

#     def subgraph_forward(self, x, g, edge_weight=None):
#         edge_index = g['edge_index']
#         if edge_weight is None:
#             # edge_weight in arguments has the highest priority
#             edge_weight = g['edge_attr']

#         for conv in self.convs:
#             # conv already implements the residual connection
#             x = conv(x, edge_index, edge_weight)

#         return x


# class EdgeGNN(NodeGNN):
#     def __init__(self, num_nodes, gcn_in_dim, gcn_dim, gcn_node_dim):
#         super().__init__(gcn_in_dim, )

#         self.gcn_dim = gcn_dim
#         self.num_nodes = num_nodes
#         self.node_dim = gcn_node_dim
#         self.edge_dim = 2*(self.gcn_dim+self.node_dim)

#         self.node_emb = nn.Embedding(self.num_nodes, self.node_dim)
#         self.edge_map = nn.Sequential(
#             nn.Linear(self.edge_dim, 1),
#             nn.ReLU(),
#         )

#     def subgraph_forward(self, x, g):
#         x = super().subgraph_forward(x, g)
#         batch_size, node_num, _ = x.shape

#         # add node-specific representations
#         n_id = g['cent_n_id']
#         x_id = self.node_emb(n_id)\
#             .reshape(1, node_num, self.node_dim)\
#             .expand(batch_size, -1, -1)
#         x = torch.cat([x, x_id], dim=-1)

#         # calculate the edge gate for each node pair
#         edge_index = g['edge_index'].permute(1, 0)  # [num_edges, 2]
#         edge_num = edge_index.shape[0]
#         edge_x = x[:, edge_index, :]\
#             .reshape(batch_size, edge_num, self.edge_dim)
#         edge_gate = self.edge_map(edge_x)

#         return edge_gate



# #-------------------------------Main Model----------------------


# class HierST(BaseModel):
#     def __init__(self, 
#                 num_nodes=None,
#                 num_features=None,
#                 num_timesteps_input=None,
#                 num_timesteps_output=None,
#                 rnn_model='nbeats',
#                 rnn_nhid = 32,
#                 use_default_edge=False,
#                 gcn_hid=32,
#                 device = 'cpu'):
#         super().__init__()
#         self.device = device
#         self.m = num_nodes
#         self.nfeat = num_features
#         self.w = num_timesteps_input
#         self.h = num_timesteps_output


#         self.rnn_type = rnn_model
#         if self.rnn_type == 'nbeats':
#             self.rnn = NBeatsEncoder(num_nodes, date_emb_dim, id_emb_dim, lookahead_days, lookback_days, hidden_dim, day_fea_dim, block_size)
#             self.rnn_hid_dim = rnn_nhid
#         else:
#             raise Exception(f'Unsupported rnn type {self.rnn_type}')

#         if use_default_edge:
#             self.edge_gnn = None
#         else:
#             self.edge_gnn = EdgeGNN(num_nodes, self.rnn_hid_dim, gcn_dim, gcn_node_dim)
#         self.node_gnn = NodeGNN(self.rnn_hid_dim, )
#         self.gnn_fc = nn.Linear(gcn_hid, )

#         self.edge_gate = None
#         self.y_t = None
#         self.y_g = None

#     def forward(self, input_day, g):
#         # rnn_out.size: [batch_size, node_num, hidden_dim]
#         # y_rnn.size: [batch_size, node_num, forecast_len]
#         if self.rnn_type == 'nbeats':
#             # nb_out.size: [batch_size, node_num, hidden_dim, seq_len]
#             nb_out, self.y_t = self.rnn(input_day, g)
#             rnn_out, _ = nb_out.max(dim=-1)
#         else:
#             raise Exception(f'Unsupported rnn type {self.rnn_type}')

#         # edge_gate.size: [batch_size, edge_num, 1]
#         # gcn_out.size: [batch_size, node_num, hidden_dim]
#         # y_gcn.size: [batch_size, node_num, forecast_len]
#         if self.edge_gnn is None:
#             gcn_out = self.node_gnn(rnn_out, g)
#         else:
#             self.edge_gate = self.edge_gnn(rnn_out, g)
#             gcn_out = self.node_gnn(rnn_out, g, edge_weight=self.edge_gate)
#         self.y_g = self.gnn_fc(gcn_out)

#         y = self.y_t + self.y_g

#         return y


import torch
from torch_geometric.data import Data
import numpy as np
import os
import urllib.request

from .base import Dataset

class UniversalDataset(Dataset):
    """
    UniversalDataset class is designed to handle various types of graph data,
    enabling operations on datasets that include features, states, dynamic graphs, and edge attributes.

    Parameters
    ----------
    x : torch.Tensor, optional
        Node features tensor of shape (num_samples, num_nodes, num_features). Represents the node features over multiple timesteps.
    states : torch.Tensor, optional
        Tensor representing various states of nodes, similar in structure to node features.
    y : torch.Tensor, optional
        Tensor representing target labels or values for each node, structured similar to node features.
    graph : torch.Tensor or scipy.sparse matrix, optional
        Static graph structure as an adjacency matrix.
    dynamic_graph : torch.Tensor, optional
        Dynamic graph information over time, providing evolving adjacency matrices.
    edge_index : torch.LongTensor, optional
        Tensor containing edge indices, typically of shape (2, num_edges), for defining which nodes are connected.
    edge_weight : torch.Tensor, optional
        Edge weights corresponding to the edge_index, providing the strength or capacity of connections.
    edge_attr : torch.Tensor, optional
        Attributes or features for each edge, aligned with the structure defined in edge_index.
    """
    def __init__(
                self,
                x=None, # timestep * Num nodes * features 
                states=None,
                y=None,
                graph=None,
                dynamic_graph=None,
                edge_index=None,
                edge_weight=None,
                edge_attr = None
                ):
        
        super().__init__()

        self.x = x # N*D; L*D; L*N*D; 
        self.y = y # N*1; L*1; L*N*1
        self.graph = graph
        self.dynamic_graph = dynamic_graph
        self.edge_index = edge_index # None; 2*Links; L*2*Links
        self.edge_weight = edge_weight # same as edge_index
        self.edge_attr = edge_attr # same as edge_index
        self.states = states # same as x
        self.output_dim = None

        self.transforms = None

        if self.y is not None:
            self.output_dim = self.y.shape
        
        if self.graph is not None and self.edge_index is None:
            sparse_adj = self.graph.to_sparse()
            self.edge_index = sparse_adj.indices()
            self.edge_weight = sparse_adj.values()
    
    def get_transformed(self, transformations=None):
        if transformations is None:
            if self.transforms is not None:
                transformations = self.transforms
            else:
                raise AttributeError("transformation does not exists!")
            
        input_data = {"features": self.x, 
                      "graph": self.graph, 
                      "dynamic_graph": self.dynamic_graph, 
                      "states": self.states
                    }
        transformed_data = self.transforms(input_data)

        return transformed_data['features'], transformed_data['graph'], transformed_data['dynamic_graph'], transformed_data['states']

        
    def __getitem__(self, index):
        batch_y = None if self.y is None else self.y[index].unsqueeze(0)
        batch_dynamic = None if self.dynamic_graph is None else self.dynamic_graph[index].unsqueeze(0)

        if self.y is not None:
            return Data(x=self.x[index], y=batch_y, edge_index=self.edge_index, edge_attr=self.edge_weight, dynamic_graph=batch_dynamic)
        else:
            return Data(x=self.x[index], edge_index=self.edge_index, edge_attr=self.edge_weight, dynamic_graph=batch_dynamic)
        
    def __len__(self):
        return len(self.x)

    def download(self):
        pass

    def save(self):
        pass

    def generate_dataset(self, X = None, Y = None, states = None, dynamic_adj = None, lookback_window_size = 1, horizon_size = 1, permute = False, feat_idx = None, target_idx = None):
        """
        Takes node features for the graph and divides them into multiple samples
        along the time-axis by sliding a window of size (num_timesteps_input+
        num_timesteps_output) across it in steps of 1.
        :param X: Node features of shape (num_vertices, num_features,
        num_timesteps)
        :return:
            - Node features divided into multiple samples. Shape is
            (num_samples, num_vertices, num_features, num_timesteps_input).
            - Node targets for the samples. Shape is
            (num_samples, num_vertices, num_features, num_timesteps_output).
        """
        if X is None:
            X = self.x
        if Y is None:
            Y = self.y

        if feat_idx is not None:
            X = X[:, :, feat_idx]

        indices = [(i, i + (lookback_window_size + horizon_size)) for i in range(X.shape[0] - (lookback_window_size + horizon_size) + 1)]
        target = []
        for i, j in indices:
            target.append(Y[i + lookback_window_size: j])
        targets = torch.stack(target)
        if permute:
            targets = targets.transpose(1,2)

        input_list = [[None, X], [None, states], [None, dynamic_adj]]
        for m, inputs in enumerate(input_list):
            if inputs[1] is not None:
                tmp = []
                for i, j in indices:
                    tmp.append(inputs[1][i: i + lookback_window_size])
                input_list[m][0] = torch.stack(tmp)
                if permute:
                    inputs[0] = inputs[0].transpose(1,2)

        if target_idx is not None:
            targets = targets[:,target_idx, :]

        return input_list[0][0], targets, input_list[1][0], input_list[2][0]

    
    def load_toy_dataset(self):
        if not os.path.exists("./datasets"):
            os.mkdir("./datasets/")
        if not os.path.exists("./datasets/features.npy"):
            print("downloading toy features")
            url_feature = "https://drive.google.com/uc?export=download&id=10VRjabU1m0pluQKOTQ-GKxF3sK9bLFYF"
            urllib.request.urlretrieve(url_feature, './datasets/features.npy')

        if not os.path.exists("./datasets/graphs.npy"):
            print("downloading toy graphs")
            url_graph = "https://drive.google.com/uc?export=download&id=10ZR4k19wdWXQdPN53Tz3QAQxBiZUXEPr"
            
            urllib.request.urlretrieve(url_graph, './datasets/graphs.npy')

        data1 = np.load(f"./datasets/graphs.npy")
        data2 = np.load(f"./datasets/features.npy",  allow_pickle= True)

        self.x = torch.FloatTensor(data2.tolist()['node'])
        self.y = torch.FloatTensor(data2.tolist()['node'])[:,:,0]
       
        self.dynamic_graph = torch.FloatTensor(data2.tolist()['od'])
        self.states = torch.FloatTensor(data2.tolist()['SIR'])

        self.graph = data1
        self.edge_index = torch.FloatTensor(data1).to_sparse_coo().indices()
        self.edge_weight = torch.FloatTensor(data1).to_sparse_coo().values()
    


    

# class SpatialDataset(Dataset):
#     def __init__(   self,
#                     x = None,
#                     states = None,
#                     y = None,
#                     adj_m=None,
#                     edge_index = None,
#                     edge_attr = None
#                 ):
        
#         super().__init__()

#         self.x = x # N*D; L*D; L*N*D; 
#         self.y = y # N*1; L*1; L*N*1
#         self.adj_m = adj_m
#         self.edge_index = edge_index # None; 2*Links; L*2*Links
#         self.edge_attr = edge_attr # same as edge_index
#         self.states = states # same as x
#         self.output_dim = None
        
        
#         assert self.x is not None, "Input should not be NoneType!"
#         if self.y is not None:
#             assert len(self.x)==len(self.y), "Input and Output dim do not match!"
#             self.output_dim = self.y.shape
            
        
#         if self.edge_index is None and self.adj_m is None:
#             raise ValueError("There is no graph in Dataset, or you may specify your graph with parameter adj_m or edge_index!")
            
#         if self.adj_m is not None and self.edge_index is not None:
#             raise ValueError("There may be conflicts between your parameter adj_m and edge_index!")
        
#         if self.adj_m is not None and self.edge_index is None:
#             rows, cols = torch.where(self.adj_m != 0)
#             weights = self.adj_m[rows, cols]

#             edge_index = torch.stack([rows.long(), cols.long()], dim=0)
#             edge_weight = weights.clone().detach()
            
#             self.edge_index, self.edge_attr = edge_index, edge_weight
            
        
#         if self.adj_m is None and self.edge_index is not None:
#             num_nodes = self.x.shape[1] # or specify?
#             if self.edge_attr is None:
#                 self.edge_attr = torch.ones(self.edge_index.shape[1], 1)
#             if len(self.edge_attr.shape) == 1:
#                 self.edge_attr = torch.reshape(self.edge_attr, (self.edge_attr.shape[0], 1))
#             adj_matrix = torch.zeros((num_nodes, num_nodes, self.edge_attr.shape[1]))
#             for i in range(self.edge_index.shape[1]):
#                 src = self.edge_index[0, i]
#                 dest = self.edge_index[1, i]
#                 adj_matrix[src, dest] = self.edge_attr[i]
#             self.adj_m = adj_matrix.squeeze()
#             #print(self.adj_m.shape)
            
        
#     def __getitem__(self, index):
#         if self.y is not None:
#             return Data(x=self.x[index], y=self.y[index], edge_index=self.edge_index, edge_attr=self.edge_attr, adj_m=self.adj_m)
#         else:
#             return Data(x=self.x[index], edge_index=self.edge_index, edge_attr=self.edge_attr, adj_m=self.adj_m)
        
    
    
#     def __len__(self):
#         return len(self.x)
        






import numpy as np
from statsmodels.tsa.statespace.varmax import VARMAX
import torch.nn.init as init
import torch

class VARMAXModel:
    """
    Vector Autoregression Moving-Average with eXogenous variables (VARMAX) Model

    Parameters
    ----------
    num_features : int
        Number of features in each timestep of the input data.
    num_timesteps_input : int
        Number of timesteps considered for each input sample.
    num_timesteps_output : int
        Number of output timesteps to predict.
    
    Returns
    -------
    torch.Tensor
        A tensor of shape (batch_size, num_timesteps_output) representing the predicted values for the future timesteps.
        Each element corresponds to a predicted value for a future timestep.

    """
    def __init__(self, num_features, num_timesteps_input, num_timesteps_output):
        self.num_features = num_features
        self.num_timesteps_input = num_timesteps_input
        self.num_timesteps_output = num_timesteps_output
        self.model = None
        self.order = (1, 0)
        self.trend = 'c'

    def fit(self, train_input, train_target, val_input=None, val_target=None, epochs=1000, batch_size=10, verbose=False, patience=100):
        """
        Parameters
        ----------
        train_input : numpy.ndarray
            The input training data array, expected to be in the shape 
            (batch_size, num_timesteps_input, num_features), where:
            - `batch_size` is the number of training samples,
            - `num_timesteps_input` is the number of timesteps used as input,
            - `num_features` is the number of features per timestep.
        train_target : numpy.ndarray
            The target training data array corresponding to the inputs, shaped 
            (batch_size, num_timesteps_output).
        val_input : numpy.ndarray, optional
            The input validation data array, following the same format as `train_input`.
        val_target : numpy.ndarray, optional
            The target validation data array, following the same format as `train_target`.
        epochs : int, optional
            The number of epochs to train the model for. Default is 1000.
        batch_size : int, optional
            The size of batches to use when training the model. Default is 10.
        verbose : bool, optional
            If True, the model will print out progress during training. Default is False.
        patience : int, optional
            Number of epochs with no improvement after which training will be stopped. Default is 100.

        Returns
        -------
        list
            A list of forecasted values for each batch in the training data, each forecast corresponding 
            to the future timesteps as defined by `num_timesteps_output`.

        Notes
        -----
        This method internally converts the input PyTorch Tensors to numpy arrays if not already provided in that format, 
        fits a VARMAX model for each subset of data, and performs forecasting. It is designed to handle smaller subsets 
        of the data (defined by `subset_samples`) to demonstrate the model's fitting capability.
        """
        train_input = train_input.numpy()
        train_target = train_target.numpy()
        val_input = val_input.numpy()
        val_target = val_target.numpy()
        
        subset_samples = 10
        train_input = train_input[:subset_samples]
        train_target = train_target[:subset_samples]
        all_forecasts = []

        for data in train_input:

            if data.shape[0] > max(self.order):
                model = VARMAX(data, order=self.order, trend=self.trend)
                results = model.fit(disp=False)
                forecast = results.forecast(steps=self.num_timesteps_output)
                average_forecast = forecast.mean(axis=1)
                all_forecasts.append(average_forecast)
            else:
                all_forecasts.append(np.array([np.nan] * self.num_timesteps_output))
        
        train_target_subset = train_target[:subset_samples]

        mse_loss = np.mean((train_target_subset - all_forecasts) ** 2)
        print("MSE Loss:", mse_loss)

        return all_forecasts

    def predict(self, feature):
        test_input = feature
        test_input = test_input.numpy()
        all_forecasts = []
        for data in test_input:
            # Check if data is sufficient for the given order
            if data.shape[0] > max(self.order):
                model = VARMAX(data, order=self.order, trend=self.trend)
                results = model.fit(disp=False)
                forecast = results.forecast(steps=self.num_timesteps_output)
                average_forecast = forecast.mean(axis=1)
                all_forecasts.append(average_forecast)
            else:
                all_forecasts.append(np.array([np.nan] * self.num_timesteps_output))

        return torch.tensor(all_forecasts)
    



import torch
import os
import matplotlib.pyplot as plt

from epilearn.models.SpatialTemporal.MepoGNN import MepoGNN
from epilearn.data import UniversalDataset
from epilearn.utils import utils

# initial settings
device = torch.device('cpu')
torch.manual_seed(7)

lookback = 13 # inputs size
horizon = 5 # predicts size

permute = False
target_feat_idx = None
target_idx = None

epochs = 15 # training epochs
batch_size = 15 # training batch size

# load toy dataset
dataset = UniversalDataset()
dataset.load_toy_dataset()

# preprocessing
features, mean, std = utils.normalize(dataset.x)
adj_norm = utils.normalize_adj(dataset.graph)
adj_dynamic_norm = utils.normalize_adj(dataset.dynamic_graph)

features = features.to(device)
adj_norm = adj_norm.to(device)
adj_dynamic_norm = adj_dynamic_norm.to(device)

# prepare datasets
train_rate = 0.6 
val_rate = 0.2

split_line1 = int(features.shape[0] * train_rate)
split_line2 = int(features.shape[0] * (train_rate + val_rate))

train_original_data = features[:split_line1, :, :]
val_original_data = features[split_line1:split_line2, :, :]
test_original_data = features[split_line2:, :, :]

train_original_states = dataset.states[:split_line1, :, :]
val_original_states = dataset.states[split_line1:split_line2, :, :]
test_original_states = dataset.states[split_line2:, :, :]

train_input, train_target, train_states, train_adj = dataset.generate_dataset(X = train_original_data, Y = train_original_data[..., 0], states = train_original_states, dynamic_adj = adj_dynamic_norm, lookback_window_size = lookback, horizon_size = horizon, permute = permute, feat_idx = target_feat_idx, target_idx = target_idx)
val_input, val_target, val_states, val_adj = dataset.generate_dataset(X = val_original_data, Y = val_original_data[..., 0], states = val_original_states, dynamic_adj = adj_dynamic_norm, lookback_window_size = lookback, horizon_size = horizon, permute = permute, feat_idx = target_feat_idx, target_idx = target_idx)
test_input, test_target, test_states, test_adj = dataset.generate_dataset(X = test_original_data, Y = test_original_data[..., 0], states = test_original_states, dynamic_adj = adj_dynamic_norm, lookback_window_size = lookback, horizon_size = horizon, permute = permute, feat_idx = target_feat_idx, target_idx = target_idx)


# dynamic graph
model = MepoGNN(
        num_nodes=adj_norm.shape[0],
        num_features=train_input.shape[3],
        num_timesteps_input=lookback,
        num_timesteps_output=horizon,
        glm_type='Dynamic',
        adapt_graph = train_adj,
        blocks=2,
        layers=3,
        device=device
        ).to(device=device)

# adaptive graph learning
# model = MepoGNN(
#         num_nodes=adj_norm.shape[0],
#         num_features=train_input.shape[3],
#         num_timesteps_input=lookback,
#         num_timesteps_output=horizon,
#         glm_type='Adaptive',
#         adapt_graph = adj_norm,
#         blocks=2,
#         layers=3,
#         device=device
#         ).to(device=device)

# (glm_type='Adaptive', num_nodes = self.params['N'], in_dim=4, blocks=2, layers=3,
#                                         adpinit=self.commuter, in_len=self.params['obs_len'],
#                                         out_len=self.params['pred_len'], dropout=0.5)


# training: dynamic graph
model.fit(
        train_input=train_input, 
        train_target=train_target,
        train_states = train_states, 
        train_graph=train_adj, 
        val_input=val_input, 
        val_target=val_target, 
        val_states=val_states,
        val_graph=val_adj, 
        loss='mse',
        verbose=True,
        batch_size=batch_size,
        epochs=epochs)



import torch

from ..visualize import plot_series
from ..utils import utils, metrics  
from .base import BaseTask

class Forecast(BaseTask):
    """
    The Forecast class extends the BaseTask class, focusing on the training and evaluation of forecast models for 
    time-series prediction tasks. It includes functionalities specific to handling time-series data, especially 
    in settings that involve spatial-temporal dynamics. The class supports model initialization, training, evaluation, 
    and preprocessing, facilitating the application of various neural network architectures and configurations.
    """
    def __init__(self, prototype = None, model = None, dataset = None, lookback = None, horizon = None, device = 'cpu'):
        super().__init__(prototype, model, dataset, lookback, horizon, device)
        self.feat_mean = 0
        self.feat_std = 1


    def train_model(self,
                    dataset=None,
                    config=None,
                    permute_dataset=False,
                    train_rate=0.6,
                    val_rate=0.2,
                    loss='mse', 
                    epochs=1000, 
                    batch_size=10,
                    lr=1e-3, 
                    region_idx=None,
                    initialize=True, 
                    verbose=False, 
                    patience=100, 
                    ):
        """
        Trains the forecast model using the provided dataset and configuration settings. It handles data splitting, model 
        initialization, and the training process, and also evaluates the model on the test set, reporting metrics such as 
        MAE and RMSE.
        """
        if config is not None:
            permute_dataset = config.permute
            train_rate = config.train_rate
            val_rate = config.val_rate
            loss = config.loss
            epochs=config.epochs
            batch_size=config.batch_size
            lr=config.lr
            initialize=config.initialize
            patience=config.patience

        if dataset is None:
            try:
                dataset = self.dataset
            except:
                raise RuntimeError("dataset not exists, please input dataset or use load_dataset() first!")
        else:
            self.dataset = dataset
        
        if not hasattr(self, "model"):
            raise RuntimeError("model not exists, please use load_model() to load model first!")

        self.train_split, self.val_split, self.test_split, ((features, norm), adj) = self.get_splits(self.dataset, train_rate, val_rate, region_idx, permute_dataset)

        try:
        # initialize model
            self.model = self.prototype(
                num_nodes=adj.shape[0],
                num_features=self.train_split[0].shape[3],
                num_timesteps_input=self.lookback,
                num_timesteps_output=self.horizon,
                
                ).to(device=self.device)
            print("spatial-temporal model loaded!")
        except:
            self.model = self.prototype( num_features=self.train_split[0].shape[2],
                                    num_timesteps_input=self.lookback,
                                    num_timesteps_output=self.horizon)

            print("temporal model loaded!")

        # train
        self.model.fit(
                train_input=self.train_split[0], 
                train_target=self.train_split[1], 
                train_states = self.train_split[2],
                train_graph=adj, 
                train_dynamic_graph=self.train_split[3],
                val_input=self.val_split[0], 
                val_target=self.val_split[1], 
                val_states=self.val_split[2],
                val_graph=adj,
                val_dynamic_graph=self.val_split[3],
                verbose=True,
                batch_size=batch_size,
                epochs=epochs,
                loss=loss)
        
        # evaluate
        self.test_graph = adj
        self.test_feature = self.test_split[0]
        self.test_target = self.test_split[1]
        self.test_states = self.test_split[2]
        self.test_dynamic_graph = self.test_split[3]

        out = self.model.predict(feature=self.test_feature, 
                                 graph=self.test_graph, 
                                 states=self.test_states, 
                                 dynamic_graph=self.test_dynamic_graph
                                 )
        if type(out) is tuple:
            out = out[0]
        preds = out.detach().cpu()*norm[0]+norm[1]
        targets = self.test_target.detach().cpu()*norm[0]+norm[1]
        # metrics
        mae = metrics.get_MAE(preds, targets)
        rmse = metrics.get_RMSE(preds, targets)
        print(f"Test MAE: {mae.item()}")
        print(f"Test RMSE: {rmse.item()}")
        
        return {"mae":mae.item(), "rmse":rmse.item()}


    def evaluate_model(self,
                    model=None,
                    config=None,
                    features=None,
                    graph=None,
                    dynamic_graph=None,
                    norm=None,
                    states=None,
                    targets=None,
                    ):
        """
        Evaluates the trained model on a new dataset or using preloaded features and graphs. It outputs prediction accuracy 
        metrics such as MAE and RMSE for the forecasted values.
        """
        if model is None:
            if not hasattr(self, "model"):
                raise RuntimeError("model not exists, please use load_model() to load model first!")
            model = self.model
        
        features = self.test_feature if features is None else features
        graph = self.test_graph if graph is None else graph
        states = self.test_states if states is None else states
        dynamic_graph = self.test_dynamic_graph if dynamic_graph is None else dynamic_graph
        targets = self.test_target if targets is None else targets
        mean = self.feat_mean[0] if norm is None else norm['mean']
        std = self.feat_std[0] if norm is None else norm['std']

        # evaluate
        out = self.model.predict(feature=features, 
                                 graph=graph, 
                                 states=states, 
                                 dynamic_graph=dynamic_graph
                                 )
        if type(out) is tuple:
            out = out[0]
        preds = out.detach().cpu()*std+mean
        targets = targets.detach().cpu()*std+mean
        # metrics
        mae = metrics.get_MAE(preds, targets)
        rmse = metrics.get_RMSE(preds, targets)
        print(f"Test MAE: {mae.item()}")
        print(f"Test RMSE: {rmse.item()}")
        
        return {"mae":mae.item(), "rmse":rmse.item()}
    


    def get_splits(self, dataset=None, train_rate=0.6, val_rate=0.2, region_idx=None, permute=False):
        """
        Splits the provided dataset into training, validation, and testing sets based on specified rates. It also handles 
        preprocessing to normalize the data and prepare it for the model.
        """
        if dataset is None:
            try:
                dataset = self.dataset
            except:
                raise RuntimeError("dataset not exists, please use load_dataset() to load dataset first!")
            
        # preprocessing
        features, adj, adj_dynamic, states = dataset.get_transformed()

        feat_mean, feat_std = dataset.transforms.feat_mean, dataset.transforms.feat_std
        self.feat_mean = feat_mean
        self.feat_std = feat_std

        features = features.to(self.device)

        split_line1 = int(features.shape[0] * train_rate)
        split_line2 = int(features.shape[0] * (train_rate + val_rate))

        self.train_original_data = features[:split_line1, ...]
        self.val_original_data = features[split_line1:split_line2, ...]
        self.test_original_data = features[split_line2:, ...]
        
        train_original_states = None
        val_original_states = None
        test_original_states = None

        if adj is not None:
            adj = adj.to(self.device)
        if adj_dynamic is not None:
            adj_dynamic = adj_dynamic.to(self.device)
        if states is not None:
            states = states.to(self.device)
            train_original_states = states[:split_line1, ...]
            val_original_states = states[split_line1:split_line2, ...]
            test_original_states = states[split_line2:, ...]


        train_input, train_target, train_states, train_adj = dataset.generate_dataset(
                                                                                        X=self.train_original_data, 
                                                                                        Y=self.train_original_data[..., 0], 
                                                                                        states=train_original_states,
                                                                                        dynamic_adj = adj_dynamic,
                                                                                        lookback_window_size=self.lookback,
                                                                                        horizon_size=self.horizon, 
                                                                                        permute=permute)
        val_input, val_target, val_states, val_adj = dataset.generate_dataset(
                                                                                X=self.val_original_data, 
                                                                                Y=self.val_original_data[..., 0], 
                                                                                states=val_original_states,
                                                                                dynamic_adj = adj_dynamic,
                                                                                lookback_window_size=self.lookback, 
                                                                                horizon_size=self.horizon, 
                                                                                permute=permute)
        test_input, test_target, test_states, test_adj = dataset.generate_dataset(
                                                                                    X=self.test_original_data, 
                                                                                    Y=self.test_original_data[..., 0], 
                                                                                    states=test_original_states,
                                                                                    dynamic_adj = adj_dynamic,
                                                                                    lookback_window_size=self.lookback, 
                                                                                    horizon_size=self.horizon, 
                                                                                    permute=permute)
        
        if region_idx is not None:
            train_input = train_input[:,:,region_idx,:]
            val_input = val_input[:,:,region_idx,:]
            test_input = test_input[:,:,region_idx,:]

            train_target = train_target[:,:,region_idx]
            val_target = val_target[:,:,region_idx]
            test_target = test_target[:,:,region_idx]

            train_states = train_states[:,:,region_idx]
            val_states = val_states[:,:,region_idx]
            test_states = test_states[:,:,region_idx]

            features = features[:,region_idx,:]

        return (train_input, train_target, train_states, train_adj), (val_input, val_target, val_states, val_adj), (test_input, test_target, test_states, test_adj), ((features, (feat_std[0], feat_mean[0])), adj)
    
    def plot_forecasts(self, dataset=None, index_range=None):
        if dataset is None:
            data = self.test_original_data
        else:
            data = dataset.x

        # save groundtruth and predictions
        predictions = torch.FloatTensor()
        groundtruth = torch.FloatTensor()
        
        with torch.no_grad():
            
            for i in range(0, len(data)-self.horizon-self.lookback, self.horizon):
                history = torch.FloatTensor(data[i:i+self.lookback])
                label = data[i+self.lookback:i+self.lookback+self.horizon]

                output = self.model(history.unsqueeze(0))
                preds = (output*self.feat_std + self.feat_mean).squeeze(0)
                label = (label*self.feat_std + self.feat_mean).squeeze(-1)

                groundtruth = torch.cat([groundtruth, label])
                predictions = torch.cat([predictions, preds])
        result = torch.stack([predictions, groundtruth], dim=1)
        result = result[index_range[0]:index_range[1]]
        plot_series(result, columns = ['prediction', 'groundtruth'])

        return predictions, groundtruth

        



import torch
import numpy as np
import torch.nn as nn

import math
import torch.fft as fft
from einops import rearrange, reduce, repeat

def accuracy(output, labels):
    """Return accuracy of output compared to labels.

    Parameters
    ----------
    output : torch.Tensor
        output from model
    labels : torch.Tensor or numpy.array
        node labels

    Returns
    -------
    float
        accuracy
    """
    if not hasattr(labels, '__len__'):
        labels = [labels]
    if type(labels) is not torch.Tensor:
        labels = torch.LongTensor(labels)
    preds = output.max(1)[1].type_as(labels)
    correct = preds.eq(labels).double()
    correct = correct.sum()
    return correct / len(labels)


def normalize(X):
    """
    Normalizes the input tensor X to have zero mean and unit standard deviation. Handles 3D tensors.
    
    Parameters
    ----------
    X : torch.Tensor
        The input tensor to be normalized.

    Returns
    -------
    torch.Tensor
        The normalized tensor.
    torch.Tensor
        The means of the input tensor.
    torch.Tensor
        The standard deviations of the input tensor.
    """
    if len(X.shape) == 3:
        means = torch.mean(X, axis=(0, 1))
        X = X - means.unsqueeze(0).unsqueeze(0)
        stds = torch.std(X, axis=(0, 1))
        X = X / stds.unsqueeze(0).unsqueeze(0)
    X[torch.where(torch.isnan(X))]=0
    return X, means, stds


def normalize_adj(Adj):
    """
    Returns the degree normalized adjacency matrix.
    
    Parameters
    ----------
    Adj : torch.Tensor or np.array
        The input adjacency matrix.

    Returns
    -------
    torch.Tensor
        The degree normalized adjacency matrix.
    """
    if Adj is None:
        return None
    try:
        Adj = Adj.numpy()
    except:
        pass

    if len(Adj.shape) > 2:
        A_wave = Adj
        for i in range(Adj.shape[0]):
            A = Adj[i].reshape(Adj.shape[1], Adj.shape[1])
            A = A + np.diag(np.ones(A.shape[0], dtype=np.float32))
            D = np.array(np.sum(A, axis=1)).reshape((-1,))
            D[D <= 10e-5] = 10e-5    # Prevent infs
            diag = np.reciprocal(np.sqrt(D))
            A_wave[i] = np.multiply(np.multiply(diag.reshape((-1, 1)), A),
                                diag.reshape((1, -1))).reshape(Adj[i].shape)

    else:
        A = Adj
        A = A + np.diag(np.ones(A.shape[0], dtype=np.float32))
        D = np.array(np.sum(A, axis=1)).reshape((-1,))
        D[D <= 10e-5] = 10e-5    # Prevent infs
        diag = np.reciprocal(np.sqrt(D))
        A_wave = np.multiply(np.multiply(diag.reshape((-1, 1)), A),
                            diag.reshape((1, -1)))
    return torch.FloatTensor(A_wave)

def diff(features):
    """
    Computes the discrete difference along the time dimension of the input tensor.
    
    Parameters
    ----------
    features : torch.Tensor
        The input feature tensor.

    Returns
    -------
    torch.Tensor
        The tensor of differences along the time dimension.
    """
    return torch.diff(features, dim=0)


def Degree_Matrix(ST_matrix):
    """
    Computes the degree matrix for a given spatio-temporal adjacency matrix.
    
    Parameters
    ----------
    ST_matrix : torch.Tensor
        The input spatio-temporal adjacency matrix.

    Returns
    -------
    torch.Tensor
        The degree matrix.
    """

    row_sum = torch.sum(ST_matrix, 0)

    ## degree matrix
    dim = len(ST_matrix)
    D_matrix = torch.zeros(dim, dim)
    for i in range(dim):
        D_matrix[i, i] = 1 / max(torch.sqrt(row_sum[i]), 1)

    return D_matrix

"""
The binary spatio-temporal adjacency matrix used in USTGCN
"""
def Static_full(n, t, A):
    """
    Constructs the full spatio-temporal adjacency matrix using the binary spatio-temporal adjacency matrix method.
    
    Parameters
    ----------
    n : int
        The dimension of the spatial adjacency matrix.
    t : int
        The length of periods.
    A : torch.Tensor
        The spatial adjacency matrix.

    Returns
    -------
    torch.Tensor
        The full spatio-temporal adjacency matrix.
    """

    I_S = torch.diag_embed(torch.ones(n))
    I_T = torch.diag_embed(torch.ones(t))

    C_S = A
    C_T = torch.tril(torch.ones(t, t), diagonal=-1)

    S = I_S + C_S
    A_ST = kronecker(C_T, S) + kronecker(I_T, C_S)

    return A_ST


"""
Use kronecker product to construct the spatio-temporal adjacency matrix
"""
def kronecker(A, B):
    """
    Constructs the spatio-temporal adjacency matrix using the Kronecker product.
    
    Parameters
    ----------
    A : torch.Tensor
        The temporal adjacency matrix.
    B : torch.Tensor
        The spatial adjacency matrix.

    Returns
    -------
    torch.Tensor
        The adjacency matrix of one space-time neighboring block.
    """
    AB = torch.einsum("ab,cd->acbd", A, B)
    AB = AB.contiguous().view(A.size(0)*B.size(0), A.size(1)*B.size(1))
    return AB

def edge_to_adj(edge_index, num_nodes):
    """
    Converts edge index representation to adjacency matrix.
    
    Parameters
    ----------
    edge_index : torch.Tensor
        The edge index tensor where each column represents an edge.
    num_nodes : int
        The number of nodes in the graph.

    Returns
    -------
    torch.Tensor
        The adjacency matrix.
    """
    adj = torch.zeros(num_nodes, num_nodes)
    adj[edge_index[0], edge_index[1]] = 1
    adj[edge_index[1], edge_index[0]] = 1
    return adj


# def tensor_to_networkx(features, graph):
#     G = nx.Graph()
#     nodes = list(range(len(graph)))
#     G.add_nodes_from(nodes)
#     edges = []
#     for i in nodes:
#         for j in nodes:
#             if graph[i, j] > 0:
#                 edges.append((i, j))
#     G.add_edges_from(edges)

#     cmap = plt.cm.get_cmap('Reds')
#     node_weights = features.view(-1).tolist()  


class moving_avg(nn.Module):
    """
    Moving average block to highlight the trend of time series. This module smooths the input time series data 
    using a moving average filter, which helps in capturing the underlying trend by averaging over a specified window.

    Parameters
    ----------
    kernel_size : int
        The size of the moving average window.
    stride : int
        The stride of the moving average window.
    """

    def __init__(self, kernel_size, stride):
        super(moving_avg, self).__init__()
        self.kernel_size = kernel_size
        self.avg = nn.AvgPool1d(kernel_size=kernel_size, stride=stride, padding=0)

    def forward(self, x):
        """
        Applies the moving average to the input tensor 'x', padding the time series at both ends to ensure
        that the moving average is computed correctly for the entire series.

        Parameters
        ----------
        x : torch.Tensor
            The input time-series data tensor, typically of shape (batch_size, time_steps, features).

        Returns
        -------
        torch.Tensor
            The tensor smoothed using the moving average, maintaining the original shape of the input.
        """
        # padding on the both ends of time series
        front = x[:, 0:1, :].repeat(1, self.kernel_size - 1 - math.floor((self.kernel_size - 1) // 2), 1)
        end = x[:, -1:, :].repeat(1, math.floor((self.kernel_size - 1) // 2), 1)
        x = torch.cat([front, x, end], dim=1)
        x = self.avg(x.permute(0, 2, 1))
        x = x.permute(0, 2, 1)
        return x


class series_decomp(nn.Module):
    """
    A PyTorch module for series decomposition using a single moving average kernel. This module decomposes
    a time series into its residual and moving average components, providing a straightforward approach to
    trend extraction.

    Parameters
    ----------
    kernel_size : int
        The kernel size to be used for the moving average calculation. The kernel size defines the window
        for the moving average.
    """
    def __init__(self, kernel_size):
        super(series_decomp, self).__init__()
        self.moving_avg = moving_avg(kernel_size, stride=1)

    def forward(self, x):
        """
        Applies the series decomposition to the input tensor 'x'. It calculates the moving average using the specified kernel size
        and computes the residual by subtracting the moving average from the original series.

        Parameters
        ----------
        x : torch.Tensor
            The input time-series data tensor, typically of shape (batch_size, time_steps).

        Returns
        -------
        tuple(torch.Tensor, torch.Tensor)
            A tuple containing:
            - The residual tensor after subtracting the moving average from the input.
            - The moving average tensor computed from the input.
        """
        moving_mean = self.moving_avg(x)
        res = x - moving_mean
        return res, moving_mean


class series_decomp_multi(nn.Module):
    """
    A PyTorch module for series decomposition using multiple moving average kernels. This module decomposes
    a time series into its residual and moving average components, leveraging multiple kernel sizes for enhanced
    flexibility and accuracy in capturing trends.

    Parameters
    ----------
    kernel_size : list of int
        List of kernel sizes to be used for moving average calculations. Each kernel size defines the window
        for the moving average.

    """

    def __init__(self, kernel_size):
        super(series_decomp_multi, self).__init__()
        self.moving_avg = [moving_avg(kernel, stride=1) for kernel in kernel_size]
        self.layer = torch.nn.Linear(1, len(kernel_size))

    def forward(self, x):
        """
        Applies the series decomposition to the input tensor 'x'. It calculates multiple moving averages
        using the initialized kernels, weights them using a linear layer followed by a softmax, and computes
        the residual by subtracting the weighted moving average from the original series.

        Parameters
        ----------
        x : torch.Tensor
            The input time-series data tensor, typically of shape (batch_size, time_steps).

        Returns
        -------
        tuple(torch.Tensor, torch.Tensor)
            A tuple containing:
            - The residual tensor after subtracting the weighted moving average from the input.
            - The weighted moving average tensor computed from the input.
        """
        moving_mean = []
        for func in self.moving_avg:
            moving_avg = func(x)
            moving_mean.append(moving_avg.unsqueeze(-1))
        moving_mean = torch.cat(moving_mean, dim=-1)
        moving_mean = torch.sum(moving_mean * nn.Softmax(-1)(self.layer(x.unsqueeze(-1))), dim=-1)
        res = x - moving_mean
        return res, moving_mean




class FourierLayer(nn.Module):
    """
    A PyTorch module that applies Fourier Transform techniques to time-series data, capable of extracting
    and manipulating frequency components. It supports both simple Fourier Transform and attention-based
    frequency analysis, allowing for the extrapolation and prediction based on frequency patterns.

    Parameters
    ----------
    pred_len : int
        The length of the sequence to predict using the frequency components.
    k : int, optional
        The number of top frequency components to retain for the analysis. Default is None.
    low_freq : int, optional
        The lowest frequency index to consider in the analysis to avoid zero-frequency (DC) components.
        Default is 1.
    output_attention : bool, optional
        Flag to determine if the layer should output the attention weights. Default is False.

    """
    def __init__(self, pred_len, k=None, low_freq=1, output_attention=False):
        super().__init__()
        # self.d_model = d_model
        self.pred_len = pred_len
        self.k = k
        self.low_freq = low_freq
        self.output_attention = output_attention

    def forward(self, x):
        """
        Executes the Fourier transform operations on the input data 'x'. Depending on the configuration, it performs
        a basic Fourier transform to analyze frequency components, selects top-k frequencies, and optionally performs
        an inverse transform to either reconstruct or predict time-series data.

        Parameters
        ----------
        x : torch.Tensor
            The input time-series data tensor with shape (batch_size, time_steps, features).

        Returns
        -------
        torch.Tensor, tuple(torch.Tensor, torch.Tensor)
            Depending on the `output_attention` setting, it returns either the transformed data tensor or
            a tuple containing the transformed data tensor and attention weights.
        """

        if self.output_attention:
            return self.dft_forward(x)

        b, t, d = x.shape
        x_freq = fft.rfft(x, dim=1)

        if t % 2 == 0:
            x_freq = x_freq[:, self.low_freq:-1]
            f = fft.rfftfreq(t)[self.low_freq:-1]
        else:
            x_freq = x_freq[:, self.low_freq:]
            f = fft.rfftfreq(t)[self.low_freq:]

        x_freq, index_tuple = self.topk_freq(x_freq)
        f = repeat(f, 'f -> b f d', b=x_freq.size(0), d=x_freq.size(2))
        f = f.to(x_freq.device)
        f = rearrange(f[index_tuple], 'b f d -> b f () d').to(x_freq.device)

        return self.extrapolate(x_freq, f, t), None

    def extrapolate(self, x_freq, f, t):
        x_freq = torch.cat([x_freq, x_freq.conj()], dim=1)
        f = torch.cat([f, -f], dim=1)
        t_val = rearrange(torch.arange(t + self.pred_len, dtype=torch.float),
                          't -> () () t ()').to(x_freq.device)

        amp = rearrange(x_freq.abs() / t, 'b f d -> b f () d')
        phase = rearrange(x_freq.angle(), 'b f d -> b f () d')

        x_time = amp * torch.cos(2 * math.pi * f * t_val + phase)

        return reduce(x_time, 'b f t d -> b t d', 'sum')

    def topk_freq(self, x_freq):
        values, indices = torch.topk(x_freq.abs(), self.k, dim=1, largest=True, sorted=True)
        mesh_a, mesh_b = torch.meshgrid(torch.arange(x_freq.size(0)), torch.arange(x_freq.size(2)))
        index_tuple = (mesh_a.unsqueeze(1), indices, mesh_b.unsqueeze(1))
        x_freq = x_freq[index_tuple]

        return x_freq, index_tuple

    def dft_forward(self, x):
        """
        Applies Discrete Fourier Transform (DFT) and its inverse (iDFT) to extract and use the most
        significant frequency components for extrapolation or reconstruction of signals. This method
        specifically computes a frequency-domain representation using the DFT, selects the most
        significant frequencies, and applies an inverse DFT to produce a predicted output.

        Parameters
        ----------
        x : torch.Tensor
            The input time-series data tensor with shape (batch_size, time_steps, features).

        Returns
        -------
        tuple(torch.Tensor, torch.Tensor)
            A tuple containing:
            - The extrapolated or reconstructed time-series data based on the significant frequency components.
            - The attention weights across the frequency domain shaped for visualization or further analysis,
            indicating the importance of each frequency component.
        
        """
        T = x.size(1)

        dft_mat = fft.fft(torch.eye(T))
        i, j = torch.meshgrid(torch.arange(self.pred_len + T), torch.arange(T))
        omega = np.exp(2 * math.pi * 1j / T)
        idft_mat = (np.power(omega, i * j) / T).cfloat()

        x_freq = torch.einsum('ft,btd->bfd', [dft_mat, x.cfloat()])

        if T % 2 == 0:
            x_freq = x_freq[:, self.low_freq:T // 2]
        else:
            x_freq = x_freq[:, self.low_freq:T // 2 + 1]

        _, indices = torch.topk(x_freq.abs(), self.k, dim=1, largest=True, sorted=True)
        indices = indices + self.low_freq
        indices = torch.cat([indices, -indices], dim=1)

        dft_mat = repeat(dft_mat, 'f t -> b f t d', b=x.shape[0], d=x.shape[-1])
        idft_mat = repeat(idft_mat, 't f -> b t f d', b=x.shape[0], d=x.shape[-1])

        mesh_a, mesh_b = torch.meshgrid(torch.arange(x.size(0)), torch.arange(x.size(2)))

        dft_mask = torch.zeros_like(dft_mat)
        dft_mask[mesh_a, indices, :, mesh_b] = 1
        dft_mat = dft_mat * dft_mask

        idft_mask = torch.zeros_like(idft_mat)
        idft_mask[mesh_a, :, indices, mesh_b] = 1
        idft_mat = idft_mat * idft_mask

        attn = torch.einsum('bofd,bftd->botd', [idft_mat, dft_mat]).real
        return torch.einsum('botd,btd->bod', [attn, x]), rearrange(attn, 'b o t d -> b d o t')



    
class series_decomp(nn.Module):
    """
    A PyTorch module for series decomposition using a single moving average kernel. This module decomposes
    a time series into its residual and moving average components, providing a straightforward approach to
    trend extraction.

    Parameters
    ----------
    kernel_size : int
        The kernel size to be used for the moving average calculation. The kernel size defines the window
        for the moving average.
    """

    def __init__(self, kernel_size):
        super(series_decomp, self).__init__()
        self.moving_avg = moving_avg(kernel_size, stride=1)

    def forward(self, x):
        """
        Applies the series decomposition to the input tensor 'x'. It calculates the moving average using the specified kernel size
        and computes the residual by subtracting the moving average from the original series.

        Parameters
        ----------
        x : torch.Tensor
            The input time-series data tensor, typically of shape (batch_size, time_steps).

        Returns
        -------
        tuple(torch.Tensor, torch.Tensor)
            A tuple containing:
            - The residual tensor after subtracting the moving average from the input.
            - The moving average tensor computed from the input.
        """
        moving_mean = self.moving_avg(x)
        res = x - moving_mean
        return res, moving_mean







from .base import BaseModel
import torch.nn as nn
from .base import BaseModel
import torch.nn.init as init
import torch

class LSTMModel(BaseModel):
    """
    Long Short-Term Memory (LSTM) Model

    Parameters
    ----------
    num_features : int
        Number of features in each timestep of the input data.
    num_timesteps_input : int
        Number of timesteps considered for each input sample.
    num_timesteps_output : int
        Number of output timesteps to predict.
    nhid : int, optional
        Number of hidden units in the LSTM layers. Default: 256.
    dropout : float, optional
        Dropout rate for regularization during training to prevent overfitting. Default: 0.5.
    use_norm : bool, optional
        Whether to apply layer normalization after the LSTM layers. Default: False.

    Returns
    -------
    torch.Tensor
        A tensor of shape (batch_size, num_timesteps_output) representing the predicted values for the future timesteps.
        This tensor is the output from the last timestep processed through a linear layer to predict the desired number of future timesteps.

    """
    def __init__(self, num_features, num_timesteps_input, num_timesteps_output, nhid=256, dropout=0.5, use_norm=False):
        super(LSTMModel, self).__init__()
        self.num_features = num_features  
        self.num_timesteps_input = num_timesteps_input  
        self.num_timesteps_output = num_timesteps_output  
        self.nhid = nhid 
        self.dropout = dropout
        self.use_norm = use_norm

        self.lstm = nn.LSTM(num_features, nhid, batch_first=True)

        if self.use_norm:
            self.norm = nn.LayerNorm(nhid)

        self.dropout_layer = nn.Dropout(dropout)

        self.out = nn.Linear(nhid, num_timesteps_output)

    def forward(self, x):
        """
        Parameters
        ----------
        x : torch.Tensor
            The input tensor for the model. Expected shape is 
            (batch_size, num_timesteps_input, num_features), where
            `batch_size` is the number of samples in the batch,
            `num_timesteps_input` is the number of input timesteps,
            and `num_features` is the number of features for each timestep.

        Returns
        -------
        torch.Tensor
            The output of the model, a tensor of shape 
            (batch_size, num_timesteps_output), representing the predicted values 
            for the future timesteps. Each element corresponds to a predicted value 
            for a future timestep.
        """
        lstm_out, _ = self.lstm(x)

        if self.use_norm:
            lstm_out = self.norm(lstm_out)

        lstm_out = self.dropout_layer(lstm_out)

        lstm_out = lstm_out[:, -1, :]

        output = self.out(lstm_out)
        return output

    def initialize(self):
        for name, param in self.lstm.named_parameters():
            if 'weight_ih' in name:
                init.xavier_uniform_(param.data)
            elif 'weight_hh' in name:
                init.orthogonal_(param.data)
            elif 'bias' in name:
                param.data.fill_(0)
        
        if self.use_norm:
            self.norm.reset_parameters()

        # Initialize the output layer
        init.xavier_uniform_(self.out.weight)
        self.out.bias.data.fill_(0)


    



# this code is adopted from https://github.com/CrickWu/DL4Epi/blob/master/models/CNNRNN_Res.py

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn import Parameter

from .base import BaseModel

class CNNRNN_Res(BaseModel):
    """
    Combined Convolutional Neural Network and Recurrent Neural Network with Residual Connections (CNNRNN_Res)

    Parameters
    ----------
    num_nodes : int
        Number of nodes in the graph.
    num_features : int
        Number of features per node per timestep.
    num_timesteps_input : int
        Number of timesteps considered for each input sample (window size).
    num_timesteps_output : int
        Number of output timesteps to predict.
    nhid : int, optional
        Number of hidden units in the GRU layer. Default: 32.
    residual_ratio : float, optional
        Proportion of the residual connection compared to the GRU output. Default: 0.
    residual_window : int, optional
        Number of timesteps to include in the residual connection. Default: 0.
    dropout : float, optional
        Dropout rate for regularization during training to prevent overfitting. Default: 0.5.
    device : str, optional
        The device (cpu or gpu) on which the model will be run. Default: 'cpu'.

    Returns
    -------
    torch.Tensor
        A tensor of shape (batch_size, num_timesteps_output, num_nodes), representing the predicted values for each node over future timesteps.
        Each slice along the second dimension corresponds to a timestep, with each column representing a node.
    """
    def __init__(self, 
                num_nodes, 
                num_features,
                num_timesteps_input,
                num_timesteps_output,
                nhid=32,
                residual_ratio=0,
                residual_window=0,
                dropout=0.5,
                device='cpu'):
        super(CNNRNN_Res, self).__init__()
        self.nfeat = num_features
        self.ratio = residual_ratio
        self.device = device
        self.P = num_timesteps_input
        self.n_out = num_timesteps_output
        self.m = num_nodes
        self.hidR = nhid
        self.GRU = nn.GRU(self.m, self.hidR, batch_first=True)
        self.residual_window = residual_window
        self.mask_mat = Parameter(torch.Tensor(self.m, self.m))

        self.squeeze_features = nn.Linear(self.nfeat, 1)

        self.dropout = nn.Dropout(p=dropout)
        self.linear = nn.Linear(self.hidR, self.m)
        if (self.residual_window > 0):
            self.residual_window = min(self.residual_window, self.P)
            self.residual = nn.Linear(self.residual_window, 1)
        
        self.out = nn.Linear(1, self.n_out)

    def forward(self, x, adj, states=None, dynamic_adj=None, **kargs):
        """
        Parameters
        ----------
        x : torch.Tensor
            Input features tensor with shape (batch_size, num_timesteps_input, num_nodes, num_features).
        adj : torch.Tensor
            Static adjacency matrix of the graph with shape (num_nodes, num_nodes).
        states : torch.Tensor, optional
            States of the nodes if available, with the same shape as x. Default: None.
        dynamic_adj : torch.Tensor, optional
            Dynamic adjacency matrix if available, with shape similar to adj but possibly varying over time. Default: None.

        Returns
        -------
        torch.Tensor
            The output tensor of shape (batch_size, num_timesteps_output, num_nodes),
            representing the predicted values for each node over the specified output timesteps.
        """
        # first transform
        masked_adj = adj * self.mask_mat
        x = torch.matmul(masked_adj, x)
        # RNN
        # r: window (self.P) x batch x #signal (m)
        r = self.squeeze_features(x).contiguous().view(-1, self.P, self.m)
        _, r = self.GRU(r)
        r = self.dropout(torch.squeeze(r, 0))

        res = self.linear(r)

        #residual
        if (self.residual_window > 0):
            z = x[:, -self.residual_window:, :]
            z = z.permute(0,2,1).contiguous().view(-1, self.residual_window)
            z = self.residual(z)
            z = z.view(-1,self.m)
            res = res * self.ratio + z

        res = F.sigmoid(self.out(res.unsqueeze(-1))).float()

        return res.transpose(2,1)
    
    def initialize(self):
        for layer in self.children():
            if hasattr(layer, 'reset_parameters'):
                layer.reset_parameters()



import torch

from ..utils import utils, metrics  
from .base import BaseTask

class Detection(BaseTask):
    """
    The Detection class is designed to handle the training and evaluation of models for detection tasks.
    It extends the BaseTask class, incorporating specific functionalities to work with spatial-temporal
    data models. This class includes methods for model training, evaluation, and data preprocessing to
    facilitate experiments with different types of neural network architectures and configurations.
    """
    def __init__(self, prototype = None, model = None, dataset = None, lookback = None, horizon = None, device = 'cpu'):
        super().__init__(prototype, model, dataset, lookback, horizon, device)

    def train_model(self,
                    dataset=None,
                    config=None,
                    permute_dataset=True,
                    train_rate=0.6,
                    val_rate=0.2,
                    loss='ce', 
                    epochs=1000, 
                    batch_size=10,
                    lr=1e-3, 
                    region_idx=None,
                    initialize=True, 
                    verbose=False, 
                    patience=100, 
                    ):
        '''
        Trains the detection model using the provided dataset and configuration settings. It handles data splitting, model initialization, and the training process, and also evaluates the model on the test set, reporting the accuracy.
        '''
        if config is not None:
            permute_dataset = config.permute
            train_rate = config.train_rate
            val_rate = config.val_rate
            loss = config.loss
            epochs=config.epochs
            batch_size=config.batch_size
            lr=config.lr
            initialize=config.initialize
            patience=config.patience

        if dataset is None:
            try:
                dataset = self.dataset
            except:
                raise RuntimeError("dataset not exists, please input dataset or use load_dataset() first!")
        else:
            self.dataset = dataset
        
        if not hasattr(self, "model"):
            raise RuntimeError("model not exists, please use load_model() to load model first!")

        train_split, val_split, test_split, ((features, norm), adj_norm) = self.get_splits(self.dataset, train_rate, val_rate, region_idx, permute_dataset)

        try:
            self.model = self.prototype(
                                        num_nodes=adj_norm.shape[0],
                                        num_features=train_split[0].shape[-1],
                                        num_timesteps_input=self.lookback,
                                        num_timesteps_output=self.horizon,
                                        device=self.device
                                        ).to(device=self.device)
            print("spatial-temporal model loaded!")
        except:
            try:
                self.model = self.prototype( 
                                        num_features=train_split[0].shape[2],
                                        num_timesteps_input=self.lookback,
                                        num_timesteps_output=self.horizon,
                                        device=self.device).to(device=self.device)
                                        
                print("temporal model loaded!")
            except:
                if len(train_split[0].shape) == 4:
                    num_features = train_split[0].shape[-1] * train_split[0].shape[-2]
                else:
                    num_features = train_split[0].shape[-1]
                self.model = self.prototype(
                                        num_features=num_features,
                                        num_classes=self.horizon,
                                        device=self.device).to(device=self.device)
                print("spatial model loaded!")

        #print(train_split[0].shape) # torch.Size([323, 47, 4])
        #print(train_split[1].shape) # torch.Size([323, 47])
        # torch.Size([323, 47, 2])
        

        # train
        self.model.fit(
                train_input=train_split[0], 
                train_target=train_split[1], 
                train_states=train_split[2],
                train_graph=adj_norm, 
                train_dynamic_graph=train_split[3],
                val_input=val_split[0], 
                val_target=val_split[1], 
                val_states=val_split[2],
                val_graph=adj_norm,
                val_dynamic_graph=val_split[3],
                verbose=True,
                batch_size=batch_size,
                epochs=epochs,
                loss=loss)
        
        # evaluate
        out = self.model.predict(feature=test_split[0], graph=adj_norm, states=test_split[2], dynamic_graph=test_split[3])
        if type(out) is tuple:
            out = out[0]
        preds = out.detach().cpu().argmax(2)
        # metrics
        acc = metrics.get_ACC(preds, test_split[1])
        print(f"Test ACC: {acc.item()}")
        
        return {'acc': acc}


    def evaluate_model(self,
                    model=None,
                    dataset=None,
                    config=None,
                    features=None,
                    graph=None,
                    dynamic_graph=None,
                    norm={"std":1, 'mean':0},
                    states=None,
                    targets=None,
                    ):
        '''
        Evaluates the trained model on a new dataset or using preloaded features and graphs. It outputs the prediction accuracy of the model.
        '''
        if model is None:
            if not hasattr(self, "model"):
                raise RuntimeError("model not exists, please use load_model() to load model first!")
            model = self.model

        # evaluate
        out = self.model.predict(feature=features, graph=graph, states=states, dynamic_graph=dynamic_graph)
        preds = out.detach().cpu().argmax(1)
        # metrics
        acc = metrics.get_ACC(preds, targets)
        print(f"Test ACC: {acc.item()}")
    

    def get_splits(self, dataset=None, train_rate=0.6, val_rate=0.2, preprocess=False, region_idx=None, permute=False):
        '''
        Splits the provided dataset into training, validation, and testing sets based on specified rates. It also handles preprocessing if necessary.
        '''
        if dataset is None:
            try:
                dataset = self.dataset
            except:
                raise RuntimeError("dataset not exists, please use load_dataset() to load dataset first!")

        # preprocessing
        features, adj, adj_dynamic, states = dataset.get_transformed()
        feat_mean, feat_std = dataset.transforms.feat_mean, dataset.transforms.feat_std

        features = features.to(self.device)
        adj = adj.to(self.device)

        if adj_dynamic is not None:
            adj_dynamic = adj_dynamic.to(self.device)
        if states is not None:
            states = states.to(self.device)

        split_line1 = int(features.shape[0] * train_rate)
        split_line2 = int(features.shape[0] * (train_rate + val_rate))

        train_feature = features[:split_line1, :, :]
        val_feature  = features[split_line1:split_line2, :, :]
        test_feature  = features[split_line2:, :, :]

        train_target = dataset.y[:split_line1, :]
        val_target = dataset.y[split_line1:split_line2, :]
        test_target = dataset.y[split_line2:, :]
        try:
            train_states = states[:split_line1, :, :]
            val_states = states[split_line1:split_line2, :, :]
            test_states = states[split_line2:, :, :]
        except:
            train_states = None
            val_states = None
            test_states = None

        if dataset.dynamic_graph is not None: # hasattr(dataset, 'dynamic_graph') and 
            train_graph = adj_dynamic[:split_line1, :, :]
            val_graph = adj_dynamic[split_line1:split_line2, :, :]
            test_graph = adj_dynamic[split_line2:, :, :]
        else:
            train_graph = None
            val_graph = None
            test_graph = None

        return (train_feature, train_target, train_states, train_graph), (val_feature, val_target, val_states, val_graph), (test_feature, test_target, test_states, test_graph), ((features, (feat_std, feat_mean)), adj)

        




import torch.optim as optim
import torch.nn as nn
import torch.nn.functional as F
from copy import deepcopy
import torch
from tqdm import tqdm
from torch_geometric.loader import DataLoader


from ...data import UniversalDataset
from ...utils import utils, metrics


class BaseModel(nn.Module):
    def __init__(self):
        super(BaseModel, self).__init__()
        self.device = "cpu"
        self.best_model = None
        self.best_output = None
        self.out_shape = None

    def fit(self, 
            train_input, 
            train_target, 
            train_states=None, 
            train_graph=None, 
            train_dynamic_graph=None,
            val_input=None, 
            val_target=None,
            val_states=None, 
            val_graph= None, 
            val_dynamic_graph=None,
            loss='mse', 
            epochs=1000, 
            batch_size=10,
            lr=1e-3, 
            weight_decay=1e-3,
            initialize=True, 
            verbose=False, 
            patience=100, 
            shuffle=False,
            **kwargs):
        if initialize:
            self.initialize()

        train_dataset = UniversalDataset(x=train_input, y=train_target, graph=train_graph, dynamic_graph=train_dynamic_graph)
        val_dataset = UniversalDataset(x=val_input, y=val_target, graph=val_graph, dynamic_graph=val_dynamic_graph)
        
        optimizer = torch.optim.Adam(self.parameters(), lr=lr, weight_decay=weight_decay)
        loss_fn = metrics.get_loss(loss)

        training_losses = []
        validation_losses = []
        early_stopping = patience
        best_val = float('inf')
        es_flag = False
        for epoch in tqdm(range(epochs)):
            loss = self.train_epoch(optimizer = optimizer, loss_fn = loss_fn, dataset=train_dataset, graph=train_graph,
                                    batch_size = batch_size, device = self.device, shuffle=shuffle)
            training_losses.append(loss)

            val_loss, output = self.evaluate(loss_fn = loss_fn, dataset=val_dataset, graph=val_graph,
                                             batch_size = batch_size, device = self.device, shuffle=shuffle)
            validation_losses.append(val_loss)

            if best_val > val_loss:
                best_epoch = epoch
                best_train = loss
                best_val = val_loss
                self.best_output = output
                best_weights = deepcopy(self.state_dict())
                self.best_model = best_weights
                patience = early_stopping
            else:
                patience -= 1


            if verbose and epoch%1 == 0:
                print(f"######### epoch:{epoch}")
                print("Training loss: {}".format(training_losses[-1]))
                print("Validation loss: {}".format(validation_losses[-1]))

            if epoch > early_stopping and patience <= 0:
                es_flag = True
                #print(f"Early stop at Epoch {epoch}")
                break

            

        if es_flag:
            print(f"Early stop at Epoch {epoch}!")
        print("\nFinal Training loss: {}".format(training_losses[-1]))
        print("Final Validation loss: {}".format(validation_losses[-1]))
        print("Best Epoch: {}".format(best_epoch))
        print("Best Training loss: {}".format(best_train))
        print("Best Validation loss: {}".format(best_val))

        self.load_state_dict(best_weights)

        
    def train_epoch(self, optimizer, loss_fn, dataset, batch_size = 1, device = 'cpu', shuffle=False, graph=None):
        """
        Trains one epoch with the given data.
        :param feature: Training features of shape (num_samples, num_nodes,
        num_timesteps_train, num_features).
        :param target: Training targets of shape (num_samples, num_nodes,
        num_timesteps_predict).
        :param batch_size: Batch size to use during training.
        :return: Average loss for this epoch.
        """
        
        train_loader = DataLoader(dataset=dataset, batch_size=batch_size, shuffle=shuffle)
        '''edge_index = dataset.edge_index.long()
        edge_weight = dataset.edge_weight.float()'''
        
        epoch_training_losses = []
        for batch_data in train_loader:              
            self.train()
            optimizer.zero_grad()
            batch_data = batch_data.to(device)
            y_batch = batch_data.y#.long()
            x_batch = batch_data.x.float()

            edge_index = batch_data.edge_index.long()
            edge_weight = batch_data.edge_attr.float()
            
            out = self.forward(x_batch, edge_index, edge_weight)
            self.out_shape = y_batch.shape[1:]  # [47, 1]
            try:
                out = torch.reshape(out, (y_batch.shape)) # y_batch.shape=[2, 47, 1]
            except:
                out = torch.reshape(out, (batch_data.batch[-1]+1, graph.shape[0], -1))
                
            #out = out.view(-1,  y_batch.shape[1], y_batch.shape[2])
            loss = loss_fn(out, y_batch)

            loss.backward()
            optimizer.step()
            epoch_training_losses.append(loss)
        return sum(epoch_training_losses)/len(epoch_training_losses)
    
    
    def evaluate(self, loss_fn, dataset, batch_size=1, device = 'cpu', shuffle=False, graph=None):
        with torch.no_grad():
            self.eval()
            val_loader = DataLoader(dataset=dataset, batch_size=batch_size, shuffle=shuffle)
            '''edge_index = dataset.edge_index
            edge_weight = dataset.edge_weight'''
            val_losses = []
            outs = []
            for batch_data in val_loader:
                batch_data = batch_data.to(device)
                y_batch = batch_data.y
                x_batch = batch_data.x

                edge_index = batch_data.edge_index.long()
                edge_weight = batch_data.edge_attr.float()
                out = self.forward(x_batch, edge_index, edge_weight)
                try:
                    out = torch.reshape(out, (y_batch.shape))
                except:
                    out = torch.reshape(out, (batch_data.batch[-1]+1, graph.shape[0], -1))
                val_loss = loss_fn(out, y_batch)

                val_losses.append(val_loss)
                outs.append(out)
            
            return sum(val_losses)/len(val_losses), torch.cat(outs, dim=0)

    def predict(self, feature, graph=None, states=None, dynamic_graph=None, batch_size=1, device = 'cpu', shuffle=False):
        """
        Returns
        -------
        torch.FloatTensor
        """
        print("\nPredicting Progress...")
        assert self.out_shape is not None, "Please fit model first!"
        self.eval()
        dataset = UniversalDataset(x=feature, graph=graph, dynamic_graph=dynamic_graph)
        test_loader = DataLoader(dataset=dataset, batch_size=batch_size, shuffle=shuffle)
        '''edge_index = dataset.edge_index
        edge_weight = dataset.edge_weight'''
        outs = []
        for batch_data in tqdm(test_loader, total=len(test_loader)):
            batch_data = batch_data.to(self.device)
            y_batch = batch_data.y
            x_batch = batch_data.x
            
            edge_index = batch_data.edge_index.long()
            edge_weight = batch_data.edge_attr.float()
            out = self.forward(x_batch, edge_index, edge_weight)
            # self.out_shape = [47, 1]
            #  [0...9]
            try:
                out = torch.reshape(out, (batch_data.batch[-1]+1, *self.out_shape))  # [2, 47, 1]
            except:
                out = torch.reshape(out, (batch_data.batch[-1]+1, graph.shape[0], -1))
                #out = torch.argmax(out, dim=-1)
            outs.append(out)
            
        return torch.cat(outs, dim=0)#.squeeze()
    



# import numpy as np
# import torch
# import torch.nn as nn
# import torch.nn.functional as F
# import matplotlib.pyplot as plt
# import time
# import os
# import pandas as pd
# import matplotlib.pyplot as plt
# import json
# import argparse
# import random 
# from torch.nn.utils.rnn import pad_sequence
# dtype = torch.float
# import itertools
# from sklearn.preprocessing import StandardScaler
# from model_utils import *
# from data_utils import * 
# import pdb 

# dtype = torch.float
# regions = all_hhs_regions  # for rnn


# class EINN(nn.Module):
#     def __init__(
#         self,
#         args: argparse.ArgumentParser,
#         ):
#         super(EINN, self).__init__()

#         torch.manual_seed(SEED)
#         random.seed(SEED)
#         np.random.seed(SEED)

#         # read testbed setup
#         self.ode_model = 'SEIRm'  
#         self.exp = args.exp
#         self.get_hyperparameters()
#         self.pred_week = args.pred_week
#         self.region = args.region
#         self.device = torch.device(args.dev)
#         self.model_name = 'EINN' + '-' + self.ode_model
#         self.start_week = EW_START_DATA
#         min_sequence_length = 20

#         start = time.time()

#         ''' Process user input regarding regions '''
#         global all_hhs_regions, regions
#         if self.region=='all':
#             regions = all_hhs_regions
#         else:
#             regions = self.region
#         if type(regions) is str:
#             regions = list([regions])

#         ''' Import data for all regions '''
#         self.initial_conditions = {}; self.end_point = {}
#         r_seqs = []  # state sequences of features
#         r_ys = []  # state targets
#         r_S_rmse = []  # state rmse
#         for region in regions:
#             X_state, y = get_state_train_data(region,self.pred_week)
#             r_seqs.append(X_state.to_numpy())
#             r_ys.append(y)
#             ''' Import RMSE calibration data '''
#             try:
#                 S_rmse = load_rmse_data(self.start_week,self.pred_week,region)
#                 # change from M to m (from cumulative to incidence mortality)
#                 S_rmse[:,4] = y.ravel()
#             except:
#                 print(region)  
#                 raise Exception(f'not {region} in ODE calibration')  
#             self.initial_conditions[region] = S_rmse[0,:]
#             self.end_point[region] = S_rmse[-1,:]
#             r_S_rmse.append(S_rmse)
#         r_seqs = np.array(r_seqs)  # shape: [regions, time, features]
#         r_ys = np.array(r_ys)  # shape: [regions, time, 1]
#         r_S_rmse = np.array(r_S_rmse)  # shape: [regions, time, 1]

#         # Normalize
#         # One scaler per state
#         seq_scalers = dict(zip(regions, [StandardScaler() for _ in range(len(regions))]))
#         ys_scalers = dict(zip(regions, [TorchStandardScaler() for _ in range(len(regions))]))
#         rmse_scalers = dict(zip(regions, [TorchStandardScaler() for _ in range(len(regions))]))
#         r_seqs_norm = []; r_ys_norm = []; r_rmse_norm = []
#         self.S_scale = {} # dictionary of scaling vectors for each region
#         for i, r in enumerate(regions):
#             r_seqs_norm.append(seq_scalers[r].fit_transform(r_seqs[i],self.device))
#             r_ys_norm.append(ys_scalers[r].fit_transform(r_ys[i],self.device))
#             r_rmse_norm.append(rmse_scalers[r].fit_transform(r_S_rmse[i],self.device))
#             # save rmse std
#             self.S_scale[r] = rmse_scalers[r].std.reshape(-1)
#         r_seqs_norm = np.array(r_seqs_norm)
#         r_ys_norm = np.array(r_ys_norm)
#         r_rmse_norm = np.array(r_rmse_norm)
#         # two of them are used during training
#         self.rmse_scalers = rmse_scalers
#         self.ys_scalers = ys_scalers

#         ''' Prepare train and validation dataset '''

#         def create_time_seq(no_sequences, sequence_length):
#             """
#                 Creates windows of fixed size
#             """
#             # convert to small sequences for training, starting with length 10
#             seqs = []
#             # starts at sequence_length and goes until the end
#             for idx in range(no_sequences):
#                 # Sequences
#                 seqs.append(torch.arange(idx,idx+sequence_length))
#             seqs = pad_sequence(seqs,batch_first=True).type(dtype)
#             return seqs
        
#         states, seqs, seqs_masks, y, y_mask, y_weights, rmse_seqs, time_seqs = [], [], [], [], [], [], [], []
#         test_states, test_seqs, test_seqs_masks, test_time_seq = [], [], [], []
#         for region, seq, ys, rmse in zip(regions, r_seqs_norm, r_ys_norm, r_rmse_norm):
#             ys_weights = np.ones((ys.shape[0],1))
#             ys_weights[-14:] *= 5 
#             seq, seq_mask, ys, ys_mask, ys_weight, rmse_seq = create_window_seqs(seq,rmse,ys,ys_weights,min_sequence_length)
#             # normal
#             states.extend([region for _ in range(seq.shape[0])])
#             seqs.append(seq)
#             seqs_masks.append(seq_mask)
#             y.append(ys)
#             y_mask.append(ys_mask)
#             # ys weights
#             y_weights.append(ys_weight.squeeze_(2))
#             rmse_seqs.append(rmse_seq)
#             # time sequences
#             time_seq = create_time_seq(seq.shape[0],min_sequence_length+WEEKS_AHEAD*DAY_WEEK_MULTIPLIER).unsqueeze(2)
#             time_seqs.append(time_seq)
#             # now fill up the test data
#             test_states.append(region)
#             test_seqs.append(seq[[-1]]); test_seqs_masks.append(seq_mask[[-1]])
#             test_time_seq.append(time_seq[[-1]])

#         # train and validation data, combine 
#         regions_train = np.array(states, dtype="str").tolist()
#         X_train = torch.cat(seqs,axis=0).float().numpy()
#         X_mask_train = torch.cat(seqs_masks,axis=0).unsqueeze(2).float().numpy()
#         y_train = torch.cat(y,axis=0).float().numpy()
#         y_mask_train = torch.cat(y_mask,axis=0).float().numpy()
#         y_weights_train = torch.cat(y_weights,axis=0).float().numpy()
#         rmse_train = torch.cat(rmse_seqs,axis=0).float().numpy()
#         time_train = torch.cat(time_seqs,axis=0).float().numpy()


#         # same for test
#         regions_test = np.array(test_states, dtype="str").tolist() 
#         X_test = torch.cat(test_seqs,axis=0).float().numpy()
#         X_mask_test = torch.cat(test_seqs_masks,axis=0).unsqueeze(2).float().numpy()
#         time_test = torch.cat(test_time_seq,axis=0).float().numpy()
#         # for scaling time module
#         self.t_min = torch.tensor(time_train.min())
#         # t_max is also useful for ode future 
#         self.t_max = torch.tensor(time_train.max())

#         # convert dataset to use in dataloader
#         train_dataset = SeqData(regions_train, X_train, X_mask_train, y_train, y_mask_train, y_weights_train, rmse_train, time_train)
#         # note: y_val, y_mask_val, y_weights_val, rmse_val, are not needed at test time
#         empty = np.zeros_like(regions_test)
#         test_dataset = SeqData(regions_test, X_test, X_mask_test, empty, empty, empty, empty, time_test)
        
#         # create dataloaders for each region
#         self.data_loaders = {}
#         for r in regions:
#             idx = torch.tensor(np.isin(train_dataset.region,r)) #== r
#             dset_train = torch.utils.data.dataset.Subset(train_dataset, np.where(idx)[0])
#             r_train_loader = torch.utils.data.DataLoader(dset_train, batch_size=self.batch_size, shuffle=True, pin_memory=True) 
#             self.data_loaders[r] = r_train_loader
#         print(time.time() - start ,' seconds')
#         # test data loader is small so we can use only one
#         self.test_loader = torch.utils.data.DataLoader(test_dataset, batch_size=self.batch_size, shuffle=True, pin_memory=True)

#         """ create set of ODEs""" 
#         df = pd.read_csv(population_path,index_col='region')
#         self.pop = {}
#         for region in regions:
#             self.pop[region] = df.at[region,'PopulationEstimate2019']

#         """ Create models, one per region """
#         self.encoder = EncoderModules(regions,X_train.shape[2],self.device)
#         self.decoder = DecoderModules(regions,self.device)
#         self.out_layer = OutputModules(regions,self.device)
#         out_layer_width = 20
#         self.time_nn_mod = \
#             time_pnn_fourier(
#                 regions=regions,
#                 scale=1,
#                 out_dim=out_layer_width,
#                 device=self.device
#             ).to(self.device)

#         self.ode = \
#             ode_modules_set(\
#                 regions=regions,
#                 init_ode_wcalibration=self.init_ode_wcalibration,
#                 initial_ode_idx=0,
#                 final_ode_idx=N_ODES_MAX,
#                 device=self.device,
#                 pop=self.pop
#             ).to(self.device)
                
#         """ paths to save model """
#         self.mod_path = './models/EINN/exp{}/{}/'.format(self.exp,self.region)
#         self.rnnode_file_name = 'einn_{}_{}'.format(self.pred_week,self.ode_model)

#         self.losses = []
#         self.ode_losses = []
#         self.data_losses = []
#         self.aux_losses = []
#         self.feat_losses = []
#         self.future_ode_losses = []
#         self.kd_target_losses = []; self.kd_emb_losses = []; self.feat_data_losses = []
#         self.gradient_losses = []
#         self.ode_param_losses = []
#         self.monotonicity_losses = []
#         self.lr_values = []; self.lr_ode_values = []; self.lr_values = []
#         self.epoch = 0
#         self.best_loss = torch.tensor(np.inf,dtype=dtype).to(self.device)
#         self.early = False
#         self.train_start_flag = []
#         self.save_load_ode_params = True
#         self.feat_mode = True


#     def save_odernn_local(self,suffix):

#         print('==== saving model ====')
#         def save_model(file_prefix: str, model: nn.Module, region: str):
#             torch.save(model.encoder.mods['encoder_'+region].state_dict(), file_prefix + "_encoder.pth")
#             torch.save(model.decoder.mods['decoder_'+region].state_dict(), file_prefix + "_decoder.pth")
#             torch.save(model.time_nn_mod.time_mods['time_nn_'+region].state_dict(), file_prefix + "_time_nn.pth")
#             if self.save_load_ode_params:
#                 torch.save(model.ode.state_dict(), file_prefix + "_ode.pth")
#             torch.save(model.out_layer.mods['output_'+region].state_dict(), file_prefix + "_out_layer.pth")

#         file_name = 'einn_{}_{}{}'.format(self.pred_week,self.ode_model,suffix)
#         for region in regions:
#             mod_path = './models/EINN/exp{}/{}/'.format(self.exp,region)
#             if not os.path.exists(mod_path):
#                 os.makedirs(mod_path)
#             save_model(mod_path+file_name,self,region)    

#         print('==== saved ====')    
        

#     def upload_odernn_local(self,suffix=''):

#         ''' this works well, saves regional model in the correct one'''

#         def load_model(file_prefix: str, model: nn.Module, region: str):
#             model.encoder.mods['encoder_'+region].out_layer[0].bias
#             model.encoder.mods['encoder_'+region].load_state_dict(torch.load(file_prefix + "_encoder.pth",map_location=self.device),strict=False)
#             model.decoder.mods['decoder_'+region].load_state_dict(torch.load(file_prefix + "_decoder.pth",map_location=self.device),strict=False)
#             model.time_nn_mod.load_state_dict(torch.load(file_prefix + "_time_nn.pth",map_location=self.device),strict=False)
#             if self.save_load_ode_params:
#                 model.ode.load_state_dict(torch.load(file_prefix + "_ode.pth",map_location=self.device),strict=False)
#             model.out_layer.mods['output_'+region].load_state_dict(torch.load(file_prefix + "_out_layer.pth",map_location=self.device),strict=False)

#         if self.pretrained == 'None':
#             experiment = self.exp
#         else:
#             experiment = self.pretrained

#         print('\n==== loading model ====')
#         file_name = 'einn_{}_{}{}'.format(self.pred_week,self.ode_model,suffix)
#         for region in regions:
#             mod_path = './models/EINN/exp{}/{}/'.format(experiment,region)
#             if not os.path.exists(mod_path):
#                 os.makedirs(mod_path)
#             load_model(mod_path+file_name,self,region)
#         print('load complete')

#     def forward_feature(self,region,X,X_mask,time_seq):
#         ''' 
#             Feature module forward pass
#         '''
#         X_embeds = self.encoder.mods['encoder_'+region[0]].forward_mask(X.transpose(1, 0), X_mask.transpose(1, 0))
#         time_seq = time_seq[:,-WEEKS_AHEAD*DAY_WEEK_MULTIPLIER:,:]
#         Hi_data = (time_seq - self.t_min)/(self.t_max - self.t_min)
#         emb_prime = self.decoder.mods['decoder_'+region[0]](Hi_data,X_embeds)
#         states_prime = self.out_layer.mods['output_'+region[0]](emb_prime) 
#         return states_prime, emb_prime

#     def forward_time(self,region,time_seq):
#         ''' 
#             Time module forward pass
#             Works for a single region at the time 
#         '''
#         inputi_data = time_seq.permute(1,0,2)
#         Hi_data = (inputi_data - self.t_min)/(self.t_max - self.t_min)
#         emb_Ei_data = self.time_nn_mod.time_mods['time_nn_'+region[0]](Hi_data)
#         statesi_data = self.out_layer.mods['output_'+region[0]](emb_Ei_data)

#         return statesi_data.permute(1,0,2), emb_Ei_data.permute(1,0,2)
        
#     def forward_ode(self,region,time_seq):
#         """
#             Compute ds/dt for time module
#         """
#         # pass w/ require grad
#         t_eqns = torch.tensor(time_seq,requires_grad=True)
#         states_eqns, _ = self.forward_time(region,t_eqns)

#         # scale to get gradients
#         states_eqns = self.rmse_scalers[region[0]].inverse_transform(states_eqns)
        
#         ones_tensor = torch.ones_like(t_eqns,device=self.device)
#         stateS_dt = torch.autograd.grad(states_eqns[:,:,[0]],t_eqns,grad_outputs=ones_tensor,create_graph=True) 
#         stateE_dt = torch.autograd.grad(states_eqns[:,:,[1]],t_eqns,grad_outputs=ones_tensor,create_graph=True) 
#         stateI_dt = torch.autograd.grad(states_eqns[:,:,[2]],t_eqns,grad_outputs=ones_tensor,create_graph=True) 
#         stateR_dt = torch.autograd.grad(states_eqns[:,:,[3]],t_eqns,grad_outputs=ones_tensor,create_graph=True) 
#         state_dt = torch.cat((stateS_dt[0], stateE_dt[0], stateI_dt[0], stateR_dt[0]), 2)
#         return states_eqns, state_dt
    
#     def forward_gradient_feat(self,region,X,X_mask,time_seq):
#         """ forward pass to calculate ds/dt for feature module """

#         """ calculate de/dt via time module """
#         t_eqns = time_seq[:,-WEEKS_AHEAD*DAY_WEEK_MULTIPLIER:]
#         inputi_eqns = torch.tensor(t_eqns.permute(1,0,2),requires_grad=True)
#         Hi_data = (inputi_eqns - self.t_min)/(self.t_max - self.t_min)
#         emb_Ei_eqns = self.time_nn_mod.time_mods['time_nn_'+region[0]](Hi_data)
#         dEdt = []
#         ones_tensor = torch.ones_like(inputi_eqns,device=self.device)
#         for i in range(emb_Ei_eqns.shape[2]): # size of hidden
#             dEdt.append(torch.autograd.grad(emb_Ei_eqns[:,:,[i]],inputi_eqns,grad_outputs=ones_tensor,create_graph=True)[0])
#         dEdt = torch.cat(dEdt, 2)
#         dEdt = dEdt.permute(1,0,2)
#         if self.detach_yaz:
#             dEdt = dEdt.detach()

#         """ calculate ds/de via feature module """
#         _, emb_E_prime = self.forward_feature(region,X,X_mask,time_seq)
#         # forward pass over the same feature module but saves grad on input
#         # should do scaling as per self.predict_feature 
#         emb_E_grad = torch.tensor(emb_E_prime,requires_grad=True)
#         states_prime = self.out_layer.mods['output_'+region[0]](emb_E_grad) 
#         # scale to get gradients
#         states_prime = self.rmse_scalers[region[0]].inverse_transform(states_prime)
#         ones_tensor = torch.ones_like(states_prime[:,:,4],device=self.device)
#         stateS_de = torch.autograd.grad(states_prime[:,:,0],emb_E_grad,grad_outputs=ones_tensor,create_graph=True)[0] 
#         stateE_de = torch.autograd.grad(states_prime[:,:,1],emb_E_grad,grad_outputs=ones_tensor,create_graph=True)[0] 
#         stateI_de = torch.autograd.grad(states_prime[:,:,2],emb_E_grad,grad_outputs=ones_tensor,create_graph=True)[0] 
#         stateR_de = torch.autograd.grad(states_prime[:,:,3],emb_E_grad,grad_outputs=ones_tensor,create_graph=True)[0] 

#         """ multiply previous two results """
#         stateS_dt = (dEdt * stateS_de).sum(2).unsqueeze(2)
#         stateE_dt = (dEdt * stateE_de).sum(2).unsqueeze(2)
#         stateI_dt = (dEdt * stateI_de).sum(2).unsqueeze(2)
#         stateR_dt = (dEdt * stateR_de).sum(2).unsqueeze(2)
#         gradient_feat = torch.cat((stateS_dt, stateE_dt, stateI_dt, stateR_dt), 2)
#         return states_prime, gradient_feat 

#     def data_loss(self,states_data,y,y_mask,y_w):
#         """ data loss for time module """
#         ys_data_mask = y_mask
#         ys_data_weights = y_w

#         total_data_target_tokens = torch.sum(ys_data_mask != 0).cpu() # denominator of loss

#         criterion = nn.MSELoss(reduction='none')
#         data_loss = \
#                 ((criterion(
#                     states_data[:,:,4],
#                     y[:,:,0]
#                     ) * ys_data_mask * ys_data_weights).sum() )/total_data_target_tokens
#         return data_loss

#     def feat_data_loss(self,states_prime,y,y_mask,y_w):
#         """ data loss for feature module """
#         ys_data_mask = y_mask
#         ys_data_weights = y_w
#         total_data_target_tokens = torch.sum(ys_data_mask != 0).cpu() # denominator of loss
#         criterion = nn.MSELoss(reduction='none')
#         feat_data_loss = (
#             criterion(
#                 states_prime[:,:,4],
#                 y[:,-WEEKS_AHEAD*DAY_WEEK_MULTIPLIER:,0]
#             ) * ys_data_mask[:,-WEEKS_AHEAD*DAY_WEEK_MULTIPLIER:] * \
#                 ys_data_weights[:,-WEEKS_AHEAD*DAY_WEEK_MULTIPLIER:]
#             ).sum()
#         feat_data_loss /= total_data_target_tokens
#         return feat_data_loss

#     def aux_loss(self,states_data,y_mask,rmse_seq):
#         """ 
#             uses analytical data as auxliary aid for learning hidden dynamics 
#         """
#         cols_select = np.array([True, True, True, True, False])
#         ys_data_mask = y_mask.unsqueeze(2).expand(-1,-1,4)
#         total_data_target_tokens = torch.sum(ys_data_mask != 0).cpu() # denominator of loss
#         aux_data = rmse_seq
#         aux_loss = (\
#             F.mse_loss(
#                 states_data[:,:,cols_select],
#                 aux_data[:,:,cols_select],
#                 reduction='none'
#                 ) * ys_data_mask
#             ).sum() / total_data_target_tokens
#         return aux_loss


#     def ode_loss(self,region,states_eqns,state_dt,time_seq,y,y_mask,y_w):
#         """
#             ODE loss for time module
#         """
#         t_eqns = time_seq
#         ys_eqns_mask = y_mask.unsqueeze(2).expand(-1,-1,5) # 4: self.no_ode_states-1
#         ys_eqns_weights = y_w.unsqueeze(2).expand(-1,-1,5)
#         total_eqns_target_tokens = torch.sum(ys_eqns_mask != 0).cpu() # denominator of loss
#         batch_size = t_eqns.shape[0]

#         # scale y
#         ys = self.ys_scalers[region[0]].inverse_transform(y)
        
#         dstate_time_nn = []
#         ode_loss = torch.tensor(0.,dtype=dtype,device=self.device)
#         # we go over each row, which represents a sequence of predictions that follow an ode
#         for rowi in range(batch_size):
#             # I need non-scaled time sequence
#             dstate = []
#             for j, t in enumerate(t_eqns[rowi,:,:]):
#                 ode_id = region[0] + ' ode_'+str(int(t.item()))
#                 dstate.append(self.ode.ode_mods[ode_id].ODE(states_eqns[rowi,[j],:]))
#             dstate = torch.cat(dstate,axis=0)
#             ode_loss_i = (
#                     F.mse_loss(torch.zeros_like(dstate[:,:4]),
#                     (dstate[:,:4]-state_dt[rowi,:,:])/self.S_scale[region[0]][:4],reduction='none') * \
#                         ys_eqns_mask[rowi,:,:4] * ys_eqns_weights[rowi,:,:4] 
#                     ).sum(0)  
#             ode_loss += ode_loss_i.sum()   
#             dstate_time_nn.append(dstate.unsqueeze(0))
#         dstate_time_nn = torch.cat(dstate_time_nn,axis=0)
        
#         # dstate is coming from ode eqns, we want to make dIM to be equal to inc deaths
#         # there is no dIM in autograd outgrad, so we don't need to do it
#         criterion = nn.MSELoss(reduction='none')
#         ode_loss += (criterion(
#             dstate_time_nn[:,:,4] /self.S_scale[region[0]][4],  # S_scale[4] comes from ground truth, do not change to S_grad_scale
#             ys[:,:,0] /self.S_scale[region[0]][4]
#             ) * ys_eqns_mask[:,:,4] * ys_eqns_weights[:,:,4] 
#             ).sum()
        
#         # finally divide by all tokens
#         ode_loss /= (total_eqns_target_tokens)  
#         return ode_loss

#     def future_ode_loss(self,region,states_eqns,state_dt,time_seq,y_mask):
#         """
#             ODE loss in extrapolation domain
#         """
#         t_eqns = time_seq
#         ys_eqns_mask = y_mask.unsqueeze(2).expand(-1,-1,5) # 4: self.no_ode_states-1
#         ys_eqns_mask = torch.ones_like(ys_eqns_mask) - ys_eqns_mask
#         total_eqns_target_tokens = torch.sum(ys_eqns_mask != 0).cpu() # denominator of loss
#         batch_size = t_eqns.shape[0]

#         # if there is no future
#         if total_eqns_target_tokens == 0:
#             return torch.tensor(0.0,dtype=dtype,device=self.device)

#         dstate_time_nn = []
#         ode_loss = torch.tensor(0.,dtype=dtype,device=self.device)
#         # we go over each row, which represents a sequence of predictions that follow an ode
#         ode_id = region[0] + ' ode_'+str(int(self.t_max.item()))
#         for rowi in range(batch_size):
#             dstate = self.ode.ode_mods[ode_id].ODE_detach(states_eqns[rowi,:])
#             ode_loss_i = (
#                     F.mse_loss(torch.zeros_like(dstate[:,:4]),
#                     (dstate[:,:4]-state_dt[rowi,:,:])/self.S_scale[region[0]][:4],reduction='none') * \
#                         ys_eqns_mask[rowi,:,:4] 
#                     ).sum(0)  # element-wise scaling
#             ode_loss += ode_loss_i.sum()   
#             dstate_time_nn.append(dstate.unsqueeze(0))
#         dstate_time_nn = torch.cat(dstate_time_nn,axis=0)

#         # dstate is coming from ode eqns, we want to make dIM to be equal to inc deaths
#         # there is no dIM in autograd outgrad, so we don't need to do it
#         criterion = nn.MSELoss(reduction='none')
#         ode_loss += (
#             criterion(
#                 dstate_time_nn[:,:,4] /self.S_scale[region[0]][4], 
#                 states_eqns[:,:,4] /self.S_scale[region[0]][4]
#             )* ys_eqns_mask[:,:,4] 
#             ).sum()
#         # finally divide by all tokens
#         ode_loss /= (total_eqns_target_tokens)  
#         return ode_loss

#     def kd_loss(self,states_data,states_prime,emb_E,emb_E_prime,y_mask,y_w):
#         """
#             KD loss for outputs and embeddings
#         """
#         if self.detach_yaz:
#             states_data = states_data[:,-WEEKS_AHEAD*DAY_WEEK_MULTIPLIER:,:].detach()
#             emb_E = emb_E[:,-WEEKS_AHEAD*DAY_WEEK_MULTIPLIER:,:].detach()
#         else:
#             states_data = states_data[:,-WEEKS_AHEAD*DAY_WEEK_MULTIPLIER:,:]  # .detach()
#             emb_E = emb_E[:,-WEEKS_AHEAD*DAY_WEEK_MULTIPLIER:,:]  # .detach()
#         ys_data_mask = y_mask[:,-WEEKS_AHEAD*DAY_WEEK_MULTIPLIER:].unsqueeze(2).expand(-1,-1,5)
#         ys_data_weights = y_w[:,-WEEKS_AHEAD*DAY_WEEK_MULTIPLIER:].unsqueeze(2).expand(-1,-1,5)
#         total_data_target_tokens = torch.sum(ys_data_mask != 0).cpu() # denominator of loss
#         criterion = nn.MSELoss(reduction='none')
#         kd_loss_target = (
#             (criterion(
#                 states_prime,  # 0 because feat module only predict that
#                 states_data
#                 ) * ys_data_mask  * ys_data_weights
#             ).sum(2) 
#             ).sum() / total_data_target_tokens
        
#         kd_loss_emb = (
#                 (criterion(
#                     emb_E_prime,
#                     emb_E
#                     ).mean(2) * ys_data_mask[:,:,4] * ys_data_weights[:,:,4] # use one column of mask as we use mean
#                 ) 
#                 ).sum() / (total_data_target_tokens/5)
#         return kd_loss_target, kd_loss_emb

    
#     def feat_ode_loss_mse(self,region,states_feat,feat_dt,time_seq,y,y_mask,y_w):
#         """
#             ODE loss for feature module
#         """
#         t_eqns = time_seq[:,-WEEKS_AHEAD*DAY_WEEK_MULTIPLIER:].detach()
#         ys_eqns_mask = y_mask[:,-WEEKS_AHEAD*DAY_WEEK_MULTIPLIER:].unsqueeze(2).expand(-1,-1,5)
#         ys_eqns_weights = y_w[:,-WEEKS_AHEAD*DAY_WEEK_MULTIPLIER:].unsqueeze(2).expand(-1,-1,5)
#         total_eqns_target_tokens = torch.sum(ys_eqns_mask != 0).cpu() # denominator of loss
#         batch_size = t_eqns.shape[0]

#         # scale y
#         y = self.ys_scalers[region[0]].inverse_transform(y)
        
#         dstate_time_nn = []
#         ode_loss = torch.tensor(0.,dtype=dtype,device=self.device)

#         # we go over each row, which represents a sequence of predictions that follow an ode
#         for rowi in range(batch_size):
#             # I need non-scaled time sequence
#             dstate = []
#             for j, t in enumerate(t_eqns[rowi,:,:]):
#                 ode_id = region[0] + ' ode_'+str(int(t.item()))
#                 dstate.append(self.ode.ode_mods[ode_id].ODE_detach(states_feat[rowi,[j],:]))
#             if self.detach_yaz:
#                 dstate = torch.cat(dstate,axis=0).detach()
#             else:
#                 dstate = torch.cat(dstate,axis=0)
#             ode_loss_i = (
#                     # F.mse_loss(torch.zeros_like(dstate),(dstate-state_dt[rowi,:,:4])/self.S_grad_scale[:4],reduction='none') * ys_eqns_mask[rowi,:].expand(self.no_ode_states-1, -1).transpose(1,0)
#                     F.mse_loss(torch.zeros_like(dstate[:,:4]),
#                     (dstate[:,:4]-feat_dt[rowi,:,:])/self.S_scale[region[0]][:4],reduction='none') * 
#                     ys_eqns_mask[rowi,:,:4] * ys_eqns_weights[rowi,:,:4] 
#                     ).sum(0) 
#             ode_loss += ode_loss_i.sum()   
#             dstate_time_nn.append(dstate.unsqueeze(0))
#         dstate_time_nn = torch.cat(dstate_time_nn,axis=0)

#         # dstate is coming from ode eqns, we want to make dIM to be equal to inc deaths
#         # there is no dIM in autograd outgrad, so we don't need to do it
#         criterion = nn.MSELoss(reduction='none')
#         ode_loss += (
#             criterion(
#                 dstate_time_nn[:,:,4] /self.S_scale[region[0]][4], 
#                 y[:,-WEEKS_AHEAD*DAY_WEEK_MULTIPLIER:,0] /self.S_scale[region[0]][4]
#             ) * ys_eqns_mask[:,:,4] * ys_eqns_weights[:,:,4] 
#             ).sum()
#         # finally divide by all tokens
#         ode_loss /= (total_eqns_target_tokens)  
#         return ode_loss

#     def future_feat_ode_loss(self,region,states_eqns,state_dt,time_seq,y_mask):
#         """
#             ODE loss in extrapolation domain
#         """

#         t_eqns = time_seq[:,-WEEKS_AHEAD*DAY_WEEK_MULTIPLIER:].detach()
#         ys_eqns_mask = y_mask[:,-WEEKS_AHEAD*DAY_WEEK_MULTIPLIER:].unsqueeze(2).expand(-1,-1,5)
#         total_eqns_target_tokens = torch.sum(ys_eqns_mask != 0).cpu() # denominator of loss
#         batch_size = t_eqns.shape[0]

#         # if there is no future
#         if total_eqns_target_tokens == 0:
#             return torch.tensor(0.0,dtype=dtype,device=self.device)

#         dstate_time_nn = []
#         ode_loss = torch.tensor(0.,dtype=dtype,device=self.device)
#         # we go over each row, which represents a sequence of predictions that follow an ode
#         ode_id = region[0] + ' ode_'+str(int(self.t_max.item()))
#         for rowi in range(batch_size):
#             dstate = self.ode.ode_mods[ode_id].ODE_detach(states_eqns[rowi,:])
#             ode_loss_i = (
#                     F.mse_loss(torch.zeros_like(dstate[:,:4]),
#                     (dstate[:,:4]-state_dt[rowi,:,:])/self.S_scale[region[0]][:4],reduction='none') * \
#                         ys_eqns_mask[rowi,:,:4] 
#                     ).sum(0)  # element-wise scaling
#             ode_loss += ode_loss_i.sum()   
#             dstate_time_nn.append(dstate.unsqueeze(0))
#         dstate_time_nn = torch.cat(dstate_time_nn,axis=0)
#         # dstate is coming from ode eqns, we want to make dIM to be equal to inc deaths
#         # there is no dIM in autograd outgrad, so we don't need to do it
#         criterion = nn.MSELoss(reduction='none')
#         ode_loss += (
#             criterion(
#                 dstate_time_nn[:,:,4] /self.S_scale[region[0]][4], 
#                 states_eqns[:,:,4] /self.S_scale[region[0]][4]
#             )* ys_eqns_mask[:,:,4] 
#             ).sum()
#         # finally divide by all tokens
#         ode_loss /= (total_eqns_target_tokens)  
#         return ode_loss

#     def ode_param_loss(self,region,time_seq,y_mask):
#         """ 
#             ODE parameter loss aims to make 
#                 these parameters change smoothly 
#         """
#         t_eqns = time_seq.detach()
#         ys_eqns_mask = y_mask
#         total_eqns_target_tokens = torch.sum(ys_eqns_mask != 0).cpu() # denominator of loss
#         batch_size = t_eqns.shape[0]
#         """
#         """
#         ode_param_loss = torch.tensor(0.,dtype=dtype,device=self.device)
#         criterion = nn.MSELoss(reduction='mean')
#         # we go over each row, which represents a sequence of predictions that follow an ode
#         for rowi in range(batch_size):
#             prev_ode_param = None
#             for j, t in enumerate(t_eqns[rowi,:,:]):
#                 # do not consider the ones that do not belong to the sequence of interest
#                 if ys_eqns_mask[rowi,j]==0:
#                     break
#                 ode_id = region[0] + ' ode_'+str(int(t.item())) 
#                 ode_param = self.ode.ode_mods[ode_id].get_param_vector()
#                 if prev_ode_param is not None:
#                     ode_param_loss += criterion(prev_ode_param,ode_param)
#                 else:
#                     prev_ode_param = ode_param
#         ode_param_loss /= total_eqns_target_tokens
#         return ode_param_loss

#     def monotonicity_loss(self,region,state_dt,y_mask):
#         """
#             calculates monotonicity loss 

#             @param state_dt: autograd derivative
#                 is coming from forward ode
#         """
#         ys_eqns_mask = y_mask
#         # consider also future:
#         ys_eqns_mask = torch.ones_like(ys_eqns_mask)
#         total_eqns_target_tokens = torch.sum(ys_eqns_mask != 0).cpu()
#         # Susceptible monotonically decreases
#         # we want ds/dt <= 0
#         # then, min ds/dt * relu(ds/dt)
#         # penalizes when ds/dt is positive
#         monotonicity_loss = state_dt[:,:,0] * F.relu(state_dt[:,:,0]) /self.S_scale[region[0]][0] * ys_eqns_mask 
#         # Recovered monotonically increases
#         # we want dR/dt >= 0
#         # then, min -dR/dt * relu(-dR/dt)
#         # penalizes when dR/dt is negative
#         monotonicity_loss += -1*state_dt[:,:,3] * F.relu(-1*state_dt[:,:,3]) /self.S_scale[region[0]][3] * ys_eqns_mask
        
#         monotonicity_loss = monotonicity_loss.sum() / (2*total_eqns_target_tokens) 
#         return monotonicity_loss

#     def set_ode_static(self):
#         for param in self.ode.parameters():
#             param.requires_grad = False

#     def set_ode_trainable(self):
#         for param in self.ode.parameters():
#             param.requires_grad = True

#     def set_all_trainable(self):
#         self.feat_nn_mod.train()
#         self.ode.train()
#         for param in self.feat_nn_mod.parameters():
#             param.requires_grad = True
#         for param in self.time_nn_mod.parameters():
#             param.requires_grad = True
#         for param in self.ode.parameters():
#             param.requires_grad = True

#     def set_yaz_static(self):
#         self.set_time_nn_static()
#         self.set_ode_static()

#     def set_time_nn_static(self):
#         self.time_nn_mod.eval()
#         for param in self.time_nn_mod.parameters():
#             param.requires_grad = False

#     def set_time_nn_trainable(self):
#         self.time_nn_mod.eval()
#         for param in self.time_nn_mod.parameters():
#             param.requires_grad = True

#     def set_out_layer_static(self):
#         self.out_layer.eval()
#         for param in self.out_layer.parameters():
#             param.requires_grad = False

#     def set_out_layer_trainable(self):
#         self.out_layer.eval()
#         for param in self.out_layer.parameters():
#             param.requires_grad = True

#     def set_feat_static(self):
#         self.encoder.eval()
#         self.decoder.eval()
#         for param in self.encoder.parameters():
#             param.requires_grad = False
#         for param in self.decoder.parameters():
#             param.requires_grad = False
#         self.rnn_static = True

#     def set_yaz_trainable(self):
#         self.time_nn_mod.train()
#         self.ode.train()
#         for param in self.time_nn_mod.parameters():
#             param.requires_grad = True
#         for param in self.ode.parameters():
#             param.requires_grad = True
#         self.yaz_static = False

#     def set_feat_trainable(self):
#         self.encoder.train()
#         self.decoder.train()
#         for param in self.encoder.parameters():
#             param.requires_grad = True
#         for param in self.decoder.parameters():
#             param.requires_grad = True
#         self.rnn_static = False

#     def train_only_rnn(self):
#         self.set_feat_static()
#         self.set_feat_trainable()

#     def get_hyperparameters(self):
#         """ using this for odenn and baselines"""
#         model_params_path = './setup/'
#         if not os.path.exists(model_params_path):
#             os.makedirs(model_params_path)
#         best_model_params_json_file = model_params_path+'EINN-params.json'
#         # read best hyperparams found for region
#         if os.path.exists(best_model_params_json_file):
#             print('test mode, using existing params json')
#             self.read_hyperparams_from_json(best_model_params_json_file) 
#         else:
#             raise Exception(f'no setup file {best_model_params_json_file}')

#     def read_hyperparams_from_json(
#         self,
#         model_params_json_file
#         ):
#         """ Reads hyperparameters for each, this are found in validation set"""

#         # read from json
#         with open(model_params_json_file) as f:
#             self.model_metadata = json.load(f)
#         #Initialize model with hyperparameters.
#         self.num_epochs = self.model_metadata['NUM_EPOCHS']
#         self.keep_training = self.model_metadata['KEEP_TRAINING']
#         self.lr = self.model_metadata['LEARNING_RATE']
#         self.loss_weights = self.model_metadata['LOSS_WEIGHTS']
#         self.init_ode_wcalibration = self.model_metadata['INIT_ODE']
#         self.batch_size = self.model_metadata['BATCH_SIZE']

#     def set_feat_flag(self):
#         # save epoch where we change to feat module
#         self.feat_flags.append(self.epoch)

#     def plot_loss(self,suffix2=''):
#         path = './figures/exp{}/'.format(self.exp)
#         if not os.path.exists(path):
#             os.makedirs(path)
#         loss_names = ['total_loss','ode_loss','f_ode_loss','data_loss','aux_loss','ode_param_loss']
#         losses = [self.losses,self.ode_losses,self.future_ode_losses,self.data_losses,self.aux_losses,self.ode_param_losses]
#         # if self.feat_mode:
#         loss_names.extend(['kd_target','kd_emb','f_data_loss','gradient_losses'])
#         losses.extend([self.kd_target_losses, self.kd_emb_losses, self.feat_data_losses, self.gradient_losses])
#         time = [i for i in range(len(losses[0]))]  
#         for loss_name, loss in zip(loss_names,losses):
#             plt.yscale('log')
#             plt.xlabel('epochs')
#             plt.plot(time,loss,label=loss_name)
#         if len(self.train_start_flag)>0:
#             for flag in self.train_start_flag:
#                 plt.axvline(x = flag,color='y',linewidth=1,linestyle="dashed")
#         plt.legend()
#         i = 0
#         if self.region=='all':
#             figname = f'losses-{self.ode_model}-{self.pred_week}-{self.exp}-{suffix2}-{i}.png'
#         else:
#             regions_str = '-'.join(self.region)
#             figname = f'{regions_str}_losses-{self.ode_model}-{self.pred_week}-{self.exp}-{suffix2}-{i}.png'
#         while os.path.exists(path+figname):
#             if self.region=='all':
#                 figname = f'losses-{self.ode_model}-{self.pred_week}-{self.exp}-{suffix2}-{i}.png'
#             else:
#                 regions_str = '-'.join(self.region)
#                 figname = f'{regions_str}_losses-{self.ode_model}-{self.pred_week}-{self.exp}-{suffix2}-{i}.png'
#             i += 1
#         plt.savefig(path+figname)
#         plt.close()

#     def save_predictions(
#         self,
#         region: str,
#         death_predictions: np.array,
#         submodule=None,
#         ):
#         """
#             Given an array w/ predictions, save as csv
#         """
#         data = np.array(
#             [
#                 np.arange(len(death_predictions))+1,
#                 death_predictions
#             ]
#         )
#         df = pd.DataFrame(data.transpose(),columns=['k_ahead','deaths'])
#         df['k_ahead'] = df['k_ahead'].astype('int8')
#         path = './results/COVID/{}/'.format(region)
#         if not os.path.exists(path):
#             os.makedirs(path)
#         model_name = 'EINN'+submodule
#         file_name = 'preds_{}_{}_exp{}.csv'.format(model_name,self.pred_week,self.exp)
#         df.to_csv(path+file_name,index=False)

#     def predict_save(self,suffix=''):
        
#         self.eval()
#         with torch.no_grad():  #saves memory
#             # only one batch
#             region, X, X_mask, _, _, _, _, time_seq = next(iter(self.test_loader)) 
#             self.eval()
#             X = X.to(self.device, non_blocking=True)
#             X_mask = X_mask.to(self.device, non_blocking=True)
#             time_seq = time_seq.to(self.device, non_blocking=True)

#             for i in range(len(region)):
#                 print('predict in ',region[i])
#                 states_prime, _ = self.forward_feature([region[i]],X[[i]],X_mask[[i]],time_seq[[i]])
#                 death_pred_feat = states_prime[:,:,4].reshape(1,-1)
#                 death_pred_feat = self.ys_scalers[region[i]].inverse_transform(death_pred_feat)
#                 death_pred_feat = death_pred_feat.reshape(-1).detach().cpu().data.numpy()
#                 self.save_predictions(region[i],death_pred_feat,''+suffix)

    
#     # https://github.com/AdityaLab/CAMul/blob/master/train_covid.py#L284
#     def evaluate(self):
#         # evaluates in training
#         total_mse_error = []
#         with torch.no_grad():
#             # go over region because architecture depends on it
#             for r in np.random.permutation(regions): 
#                 region, X, X_mask, y, y_mask, _, _, time_seq = next(iter(self.data_loaders[r])) 
#                 self.eval()
#                 X = X.to(self.device, non_blocking=True)
#                 X_mask = X_mask.to(self.device, non_blocking=True)
#                 time_seq = time_seq.to(self.device, non_blocking=True)
#                 y = y.to(self.device, non_blocking=True)
#                 y_mask = y_mask.to(self.device, non_blocking=True)

#                 # forward feature module 
#                 states_prime, _ = self.forward_feature(region,X,X_mask,time_seq)

#                 ys_data_mask = y_mask
#                 total_data_target_tokens = torch.sum(ys_data_mask != 0).cpu() # denominator of loss
#                 criterion = nn.MSELoss(reduction='none')
#                 mse_error = (
#                     criterion(
#                         states_prime[:,:,4],
#                         y[:,-WEEKS_AHEAD*DAY_WEEK_MULTIPLIER:,0]
#                     ) * ys_data_mask[:,-WEEKS_AHEAD*DAY_WEEK_MULTIPLIER:]
#                     ).sum()
                
#                 mse_error /= total_data_target_tokens
#                 total_mse_error.append(mse_error.cpu().item())

#         rmse_error = np.sqrt(np.array(total_mse_error).mean())
#         return rmse_error


#     def _train(self,epochs,time_reps,time_ode_reps,feat_time_reps,out_reps):

#         self.train_start_flag.append(self.epoch)  # save epoch where we start this training, used in loss plot

#         ############## optimizers #############
#         # time nn only
#         time_params = itertools.chain(self.time_nn_mod.parameters(),self.out_layer.parameters()) 
#         optimizer_time = torch.optim.Adam(time_params, lr=self.lr, amsgrad=True)
#         # time nn + ode only
#         time_ode_params = itertools.chain(self.time_nn_mod.parameters(),self.ode.parameters(),\
#             self.out_layer.parameters())  
#         optimizer_time_ode = torch.optim.Adam(time_ode_params, lr=self.lr, amsgrad=True)
#         # time nn + ode + feature
#         time_feat_params = itertools.chain(self.encoder.parameters(),self.decoder.parameters(),\
#             self.out_layer.parameters(),self.time_nn_mod.parameters())  
#         optimizer_time_feat = torch.optim.Adam(time_feat_params, lr=self.lr, amsgrad=True)
#         # output layers + ode params
#         out_ode_params = itertools.chain(self.ode.parameters(),self.out_layer.parameters()) 
#         optimizer_out_ode = torch.optim.Adam(out_ode_params, lr=self.lr)
#         # will search for 1% improvement at least
#         es = EarlyStopping(patience=20, min_delta=10, percentage=True)  # patience was 25

#         self.epoch = 0
#         self.train_start_flag.append(self.epoch)  # save epoch where we start this training, used in loss plot
         
#         self.best_loss = 10e3


#         rmse_val = self.evaluate()
#         print('initial validation rmse', rmse_val)
#         for epoch in range(epochs):

#             """ time solo """
#             for _ in range(time_reps):
#                 self.train_time = True
#                 self.train_feat = False
#                 self.use_ode_loss = False
#                 self.add_grad_loss = False
#                 self.detach_yaz = True  # True: detach both time and ode
#                 self.set_ode_static()
#                 self.set_feat_static()
#                 self.set_time_nn_trainable()
#                 self.set_out_layer_trainable()
#                 self.minibatch_train(optimizer_time,time_params)

#             """ time + ode """
#             for _ in range(time_ode_reps):
#                 self.train_yaz = True
#                 self.train_feat = False
#                 self.use_ode_loss = True
#                 self.add_grad_loss = False
#                 self.detach_yaz = True  # True: detach both time and ode
#                 self.set_feat_static()
#                 self.set_time_nn_trainable()
#                 self.set_ode_trainable()
#                 self.set_out_layer_trainable()
#                 self.minibatch_train(optimizer_time_ode,time_ode_params)
           
#             """ feat and time jointly """
#             for _ in range(feat_time_reps):
#                 self.train_time = True
#                 self.train_feat = True
#                 self.use_ode_loss = True
#                 self.add_grad_loss = False
#                 self.detach_yaz = False  # True: detach both time and ode
#                 self.set_ode_trainable()
#                 self.set_feat_trainable()
#                 self.set_time_nn_trainable()
#                 self.set_out_layer_trainable()
#                 self.minibatch_train(optimizer_time_feat,time_feat_params)

#             """ only out layer + ode params """
#             for _ in range(out_reps):
#                 self.train_time = True
#                 self.train_feat = True
#                 self.use_ode_loss = True
#                 self.add_grad_loss = True
#                 self.detach_yaz = False  # True: detach both time and ode
#                 self.set_time_nn_static()
#                 self.set_ode_trainable()
#                 self.set_feat_static()
#                 self.set_out_layer_trainable()
#                 self.minibatch_train(optimizer_out_ode,out_ode_params)

#             ''' save model '''
#             rmse_val = self.evaluate()
#             save_model = True

#             print(f'Train RMSE: {rmse_val}, bad: {es.num_bad_epochs}' )
            
#             if rmse_val < self.best_loss:
#                 print(f'>> best updated {self.exp} on {self.pred_week}, epoch {epoch}')
#                 self.best_loss = torch.tensor(rmse_val)
#                 if save_model:
#                     self.save_odernn_local(suffix='-post')
#             ''' early stopping ''' 
#             if self.early:
#                 if self.epoch > self.patience_before_es:
#                     if es.step(self.best_loss):
#                         print('====BREAK, early stopping ===')
#                         break 


#     def minibatch_train(self,optims,params,verbose=True):

#         self.epoch += 1

#         epoch_total_loss = []
#         epoch_total_loss = [] 
#         epoch_ode_loss = [] 
#         epoch_future_ode_loss = [] 
#         epoch_data_loss = [] 
#         epoch_kd_target_loss = [] 
#         epoch_kd_emb_loss = [] 
#         epoch_feat_data_loss = [] 
#         epoch_ode_param_loss = [] 
#         epoch_monotonicity_loss = [] 
#         epoch_feat_ode_loss = [] 
#         epoch_future_feat_ode_loss = [] 
#         epoch_aux_loss = [] 
#         i = 0
#         self.train()
#         start_time = time.time()
#         optims.zero_grad(set_to_none=True)
#         for r in np.random.permutation(regions):  # one region at the time
#             # backprop = False
#             region, X, X_mask, y, y_mask, y_w, rmse_seq, time_seq = next(iter(self.data_loaders[r])) 
            
#             X = X.to(self.device, non_blocking=True)
#             X_mask = X_mask.to(self.device, non_blocking=True)
#             time_seq = time_seq.to(self.device, non_blocking=True)
#             y = y.to(self.device, non_blocking=True)
#             y_mask = y_mask.to(self.device, non_blocking=True)
#             y_w = y_w.to(self.device, non_blocking=True)
#             rmse_seq = rmse_seq.to(self.device, non_blocking=True)

#             # forward feature module 
#             states_data, emb = self.forward_time(region,time_seq)

#             if self.train_time:
                
#                 data_loss = self.data_loss(states_data,y,y_mask,y_w)
#                 aux_loss = self.aux_loss(states_data,y_mask,rmse_seq)

#                  # sample time sequence
#                 idx = np.random.choice(np.arange(time_seq.shape[0]),3)
#                 eqns_time_seq, eqns_y, eqns_y_mask, eqns_y_w = time_seq[idx], y[idx],y_mask[idx],y_w[idx]
#                 states_eqns, state_dt = self.forward_ode(region,eqns_time_seq)
#                 ''' ODE loss '''
#                 if self.use_ode_loss:
#                     # note: states_eqns is scaled to real dimensions
#                     ode_loss = self.ode_loss(region,states_eqns,state_dt,eqns_time_seq,eqns_y,eqns_y_mask,eqns_y_w)
#                     future_ode_loss = self.future_ode_loss(region,states_eqns,state_dt,eqns_time_seq,eqns_y_mask)
#                     ''' Param consistency '''
#                     ode_param_loss = self.ode_param_loss(region,eqns_time_seq,eqns_y_mask)
#                 else:
#                     ode_loss = torch.tensor(0.0,dtype=dtype,device=self.device)
#                     future_ode_loss = torch.tensor(0.0,dtype=dtype,device=self.device)
#                     ode_param_loss = torch.tensor(0.0,dtype=dtype,device=self.device)
                
#                 ''' Monotonicity loss '''
#                 monotonicity_loss = self.monotonicity_loss(region,state_dt,eqns_y_mask)
                
#                 total_loss = self.loss_weights['data']*data_loss + \
#                         self.loss_weights['aux']*aux_loss +\
#                         self.loss_weights['ode']*ode_loss +\
#                         self.loss_weights['future_ode']*future_ode_loss +\
#                         self.loss_weights['ode_param']*ode_param_loss +\
#                             self.loss_weights['monotonicity']*monotonicity_loss
#             else:
#                 data_loss = torch.tensor(0.0,dtype=dtype,device=self.device)
#                 aux_loss = torch.tensor(0.0,dtype=dtype,device=self.device)
#                 monotonicity_loss = torch.tensor(0.0,dtype=dtype,device=self.device)
#                 ode_loss = torch.tensor(0.0,dtype=dtype,device=self.device)
#                 future_ode_loss = torch.tensor(0.0,dtype=dtype,device=self.device)
#                 ode_param_loss = torch.tensor(0.0,dtype=dtype,device=self.device)
#                 total_loss = torch.tensor(0.0,dtype=dtype,device=self.device)

#             ''' KD loss '''
#             if self.train_feat:
#                 # forward feature module 
#                 states_prime, emb_prime = self.forward_feature(region,X,X_mask,time_seq)
#                 feat_data_loss = self.feat_data_loss(states_prime,y,y_mask,y_w)
#                 kd_loss_target, kd_loss_emb = self.kd_loss(states_data,states_prime,emb,emb_prime,y_mask,y_w)
#                 kd_loss_emb = torch.tensor(0.0,dtype=dtype,device=self.device)

#                 total_loss += self.loss_weights['feat_data']*feat_data_loss +\
#                             self.loss_weights['kd_target']*kd_loss_target +\
#                                 self.loss_weights['kd_emb']* kd_loss_emb 
#             else:
#                 feat_data_loss = torch.tensor(0.0,dtype=dtype,device=self.device)
#                 kd_loss_target = torch.tensor(0.0,dtype=dtype,device=self.device)
#                 kd_loss_emb = torch.tensor(0.0,dtype=dtype,device=self.device)

#             ''' Feat ODE loss'''
#             ''' aka feat grad matching '''
#             if self.add_grad_loss:  # use it when embedding loss is small
#                 idx = np.random.choice(np.arange(time_seq.shape[0]),16)
#                 grad_time_seq, grad_X, grad_X_mask = time_seq[idx],X[idx],X_mask[idx]
#                 grad_y, grad_y_mask, grad_y_w =y[idx],y_mask[idx],y_w[idx]
#                 states_feat, feat_dt = self.forward_gradient_feat(region,grad_X,grad_X_mask,grad_time_seq)
#                 feat_ode_loss = self.feat_ode_loss_mse(region,states_feat,feat_dt,grad_time_seq,grad_y,grad_y_mask,grad_y_w)
#                 future_feat_ode_loss = \
#                     self.future_feat_ode_loss(region,states_feat,feat_dt,grad_time_seq,grad_y_mask)
#                 total_loss += self.loss_weights['gradient_loss']*feat_ode_loss +\
#                             self.loss_weights['gradient_loss']*future_feat_ode_loss
#             else:
#                 feat_ode_loss = torch.tensor(0.0,dtype=dtype,device=self.device)
#                 future_feat_ode_loss = torch.tensor(0.0,dtype=dtype,device=self.device)
#             total_loss.backward(retain_graph=False)
#             torch.nn.utils.clip_grad_norm_(params, 10)
#             optims.step()
#             optims.zero_grad(set_to_none=True)

#             epoch_total_loss.append(total_loss.detach().cpu().item())
#             epoch_ode_loss.append(ode_loss.detach().cpu().item())
#             epoch_future_ode_loss.append(future_ode_loss.detach().cpu().item())
#             epoch_data_loss.append(data_loss.detach().cpu().item())
#             epoch_kd_target_loss.append(kd_loss_target.detach().cpu().item())
#             epoch_kd_emb_loss.append(kd_loss_emb.detach().cpu().item())
#             epoch_feat_data_loss.append(feat_data_loss.detach().cpu().item())
#             epoch_ode_param_loss.append(ode_param_loss.detach().cpu().item())
#             epoch_monotonicity_loss.append(monotonicity_loss.detach().cpu().item())
#             epoch_feat_ode_loss.append(feat_ode_loss.detach().cpu().item())
#             epoch_future_feat_ode_loss.append(future_feat_ode_loss.detach().cpu().item())
#             epoch_aux_loss.append(aux_loss.detach().cpu().item())


#         epoch_total_loss = np.array(epoch_total_loss).mean()
#         epoch_ode_loss = np.array(epoch_ode_loss).mean()
#         epoch_future_ode_loss = np.array(epoch_future_ode_loss).mean()
#         epoch_data_loss = np.array(epoch_data_loss).mean()
#         epoch_kd_target_loss = np.array(epoch_kd_target_loss).mean()
#         epoch_kd_emb_loss = np.array(epoch_kd_emb_loss).mean()
#         epoch_feat_data_loss = np.array(epoch_feat_data_loss).mean()
#         epoch_ode_param_loss = np.array(epoch_ode_param_loss).mean()
#         epoch_monotonicity_loss = np.array(epoch_monotonicity_loss).mean()
#         epoch_feat_ode_loss = np.array(epoch_feat_ode_loss).mean()
#         epoch_future_feat_ode_loss = np.array(epoch_future_feat_ode_loss).mean()
#         epoch_aux_loss = np.array(epoch_aux_loss).mean()

#         elapsed = time.time() - start_time

#         if verbose:
#             print('Epoch: %d, Total: %.2e, Data-T: %.2e, ODE: %.2e, ODE-Fut: %.2e, Data-F: %.2e, ODE-F: %.2e, ODE-F-Fut: %.2e, KD-T: %.2e, KD-E: %.2e, Aux: %.2e, Mono: %.2e, Time: %.3f'
#                     %(self.epoch, epoch_total_loss.item(), epoch_data_loss.item(), epoch_ode_loss.item(), epoch_future_ode_loss.item(), \
#                         epoch_feat_data_loss.item(), epoch_feat_ode_loss.item(), epoch_future_feat_ode_loss.item(),  epoch_kd_target_loss.item(), epoch_kd_emb_loss.item(), epoch_aux_loss.item(), \
#                             epoch_monotonicity_loss.item(), elapsed))

#         ''' save losses '''
#         self.losses.append(epoch_total_loss)
#         self.ode_losses.append(epoch_ode_loss)
#         self.data_losses.append(epoch_data_loss)
#         self.kd_target_losses.append(epoch_kd_target_loss)
#         self.kd_emb_losses.append(epoch_kd_emb_loss)
#         self.feat_data_losses.append(epoch_feat_data_loss)
#         self.ode_param_losses.append(epoch_ode_param_loss)
#         self.monotonicity_losses.append(epoch_monotonicity_loss)
#         self.future_ode_losses.append(epoch_future_ode_loss)
#         self.gradient_losses.append(epoch_feat_ode_loss)
#         self.aux_losses.append(epoch_aux_loss)


#     def is_week(self,week):
#         return self.pred_week == week

#     def is_feat_mod_available(self):
#         return os.path.exists(self.mod_path+self.rnnode_file_name + "_encoder.pth")

#     def train_predict(self):

#         # check if feat module is already there
#         try:
#             if self.keep_training:
#                 pass
#             else:
#                 self.predict_save()
#                 return 
#         except:
#             raise Exception('no model loaded')

#         """  train time nn + ode """
#         self.train_time_ode()

#         """  train time + ode + feat """
#         self.train_time_ode_feat()
        
#         """  add ODE-F and train output layer """
#         self.train_out_layers()

#         """ save predictions """
#         self.predict_save()
#         # plot training losses
#         self.plot_loss()
#         return 

#     def train_time_ode(self):
#         ''' alternate training avoids messing up the ODE initialization '''

#         print('\n====== train time nn + ode =======')
#         EPOCHS = self.num_epochs
#         self._train(epochs=EPOCHS,time_reps=1,time_ode_reps=1,feat_time_reps=0,out_reps=0)  
#         return 

#     def train_time_ode_feat(self):
#         ''' resets ode, trains time + feature'''
        
#         print('\n====== train time + ode + feat =======')
#         EPOCHS = self.num_epochs
#         self._train(epochs=EPOCHS,time_reps=0,time_ode_reps=0,feat_time_reps=1,out_reps=0)  
#         return 

#     def train_out_layers(self):

#         print('\n====== train out layer + ode params =======')
#         EPOCHS = self.num_epochs  
#         self._train(epochs=EPOCHS,time_reps=0,time_ode_reps=0,feat_time_reps=0,out_reps=1)  

#         return 